/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is PSU Calendar Server Code, released March 3rd, 2003.
 *
 * The Initial Developer of the Original Code is
 * Penn State CSE 420w Mozilla Project
 * Portions created by the Initial Developer are Copyright (C) 2003
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):  Daniel Bungert <drb@debian.org>
 *                  Thomas Keiser <tkeiser@cede.psu.edu>
 *                  Benjamin Nave <bnave@psu.edu>
 *                  John Pendzick <jnp117@psu.edu>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

#include "psuCalServerSunCDEImpl.h"
#include "nsMemory.h"
#include "nsCOMPtr.h"
#include "nsError.h"
#include "nsString.h"
#include "plstr.h"

#include "psuICalServerError.h"
#include "nsComponentManagerUtils.h"
#include <stdlib.h>
#include "psuICalServer.h"

/* Implementation file */
NS_IMPL_ISUPPORTS2(psuCalServerSunCDEImpl, psuICalServer, oeIICal)

psuCalServerSunCDEImpl::psuCalServerSunCDEImpl()
{
  csaHandle = nsnull;
  hostName = nsnull;
  calendarName = nsnull;
  user = nsnull;
  passwd = nsnull;
  connectionState = PSUCAL_STATUS_CLOSED;
  serveraddr = nsnull;
  m_batchMode = PR_FALSE;
  m_eventList = new EventList;
  csaSessionState = CSA_SESSION_NONE;
}

psuCalServerSunCDEImpl::~psuCalServerSunCDEImpl()
{
  switch( connectionState )
    {
    case PSUCAL_STATUS_OPEN:
      Close( );
      /* lack of break is deliberate */
    case PSUCAL_STATUS_AUTHENTICATED:
    case PSUCAL_STATUS_CONNECTED:
      Disconnect( );
      break;
    default:
      break;
    }

  if (hostName != nsnull)
    {
      nsMemory::Free(hostName);
      hostName = nsnull;
    }
  if (calendarName != nsnull)
    {
      nsMemory::Free(calendarName);
      calendarName = nsnull;
    }
  if (user != nsnull)
    {
      nsMemory::Free(user);
      calendarName = nsnull;
    }
  if (passwd != nsnull)
    {
      nsMemory::Free(passwd);
      passwd = nsnull;
    }
  if (serveraddr != nsnull)
    { 
      nsMemory::Free(serveraddr);
      serveraddr = nsnull;
    }
  if (m_eventList != nsnull)
    {
      delete m_eventList;
      m_eventList = nsnull;
    }
}

/* readonly attribute string hostName; */
NS_IMETHODIMP psuCalServerSunCDEImpl::GetHostName(char * *aHostName)
{
#ifdef PSUCAL_DEBUG
  printf( "psuCalServerSunCDEImpl::GetHostName()\n" );
#endif
  if (aHostName == nsnull)
    {
      RETURN( NS_ERROR_NULL_POINTER );
    }
	
  *aHostName = NS_STATIC_CAST(char *, nsMemory::Clone( hostName, strlen(hostName)+1 ));
  if( *aHostName == nsnull )
    {
      RETURN( NS_ERROR_OUT_OF_MEMORY );
    }
  RETURN( NS_OK );
}

/* readonly attribute string calendarName; */
NS_IMETHODIMP psuCalServerSunCDEImpl::GetCalendarName(char * *aCalendarName)
{
#ifdef PSUCAL_DEBUG
  printf( "psuCalServerSunCDEImpl::GetCalendarName()\n" );
#endif
  if (aCalendarName == nsnull)
    {
      RETURN( NS_ERROR_NULL_POINTER );
    }
	
  *aCalendarName = NS_STATIC_CAST(char*, nsMemory::Clone( calendarName, strlen(calendarName)+1 ));
  if( *aCalendarName == nsnull )
    {
      RETURN( NS_ERROR_OUT_OF_MEMORY );
    }
  RETURN( NS_OK );
}

/* readonly attribute serverStatusProperty connectionState; */
NS_IMETHODIMP psuCalServerSunCDEImpl::GetConnectionStatus(serverStatusProperty *aConnectionStatus)
{
#ifdef PSUCAL_DEBUG
  printf( "psuCalServerSunCDEImpl::GetConnectionStatus()\n" );
#endif
  if (aConnectionStatus == nsnull)
    {
      RETURN( NS_ERROR_NULL_POINTER );
    }

  *aConnectionStatus = connectionState;
  RETURN( NS_OK );
}

/* void connect (); */
NS_IMETHODIMP psuCalServerSunCDEImpl::Connect()
{
#ifdef PSUCAL_DEBUG
  printf( "psuCalServerSunCDEImpl::Connect()\n" );
#endif
  if (connectionState != PSUCAL_STATUS_CLOSED)
    {
      RETURN( NS_ERROR_CAL_SERVER_ALREADY_CONNECTED );
    }

  connectionState = PSUCAL_STATUS_CONNECTED;
  RETURN( NS_OK );
}

/* void disconnect (); */
NS_IMETHODIMP psuCalServerSunCDEImpl::Disconnect()
{
#ifdef PSUCAL_DEBUG
  printf( "psuCalServerSunCDEImpl::Disconnect()\n" );
#endif

  CSA_return_code ret;
  nsresult rv = NS_OK;

  if (connectionState == PSUCAL_STATUS_CLOSED)
    {
      RETURN( NS_ERROR_CAL_SERVER_NOT_CONNECTED );
    }
  if (connectionState == PSUCAL_STATUS_OPEN)
    {
      RETURN( NS_ERROR_CAL_SERVER_ALREADY_OPEN );
    }

  if (csaSessionState != CSA_SESSION_NONE)
    {
      ret = csa_logoff( csaHandle, NULL );
      rv = CSAReturnToNSResult( ret );
    }

  if (NS_SUCCEEDED(rv))
    {
      connectionState = PSUCAL_STATUS_CLOSED;
      csaSessionState = CSA_SESSION_NONE;
      csaHandle = nsnull;
    }

  RETURN( rv );
}

/* void authenticate (in string user, in string pass); */
NS_IMETHODIMP psuCalServerSunCDEImpl::Authenticate(const char *user, const char *pass)
{
#ifdef PSUCAL_DEBUG
  printf( "psuCalServerSunCDEImpl::Authenticate()\n" );
#endif

  nsresult rv;

  switch ( connectionState )
    {
    case PSUCAL_STATUS_OPEN:
    case PSUCAL_STATUS_AUTHENTICATED:
      RETURN( NS_ERROR_CAL_SERVER_ALREADY_AUTHENTICATED );
      break;
    case PSUCAL_STATUS_CLOSED:
      RETURN( NS_ERROR_CAL_SERVER_NOT_CONNECTED );
      break;
    default:
      break;
    }

  connectionState = PSUCAL_STATUS_AUTHENTICATED;

  RETURN( NS_OK );
}


/* void open (); */
NS_IMETHODIMP psuCalServerSunCDEImpl::Open()
{
#ifdef PSUCAL_DEBUG
  printf( "psuCalServerSunCDEImpl::Open()\n" );
#endif

  nsresult rv;

  switch( connectionState )
    {
    case PSUCAL_STATUS_OPEN:
      RETURN( NS_ERROR_CAL_SERVER_ALREADY_OPEN );
      break;
    case PSUCAL_STATUS_CLOSED:
    case PSUCAL_STATUS_CONNECTED:
      RETURN( NS_ERROR_CAL_SERVER_NOT_AUTHENTICATED );
      break;
    default:
      break;
    }

  rv = CSALogon( );

  if (NS_FAILED(rv))
    {
      /* we can trap the calendar doesn't exist case here */
      if (rv != NS_ERROR_CAL_SERVER_CALENDAR_NOT_EXIST)
	{
	  RETURN( rv );
	}

      rv = AddCalToServer( );
      
      /* Concurrency issues mean we should ignore a "calendar already exists"
	 error.  If the server state is really messed up, the next call
	 to CSALogon will fail and give the user back an error code */
      if (NS_FAILED(rv) && (rv != NS_ERROR_CAL_SERVER_CALENDAR_EXISTS))
	{
	  RETURN( rv );
	}

      rv = CSALogon( );
    }

  if (NS_SUCCEEDED(rv))
    {
      connectionState = PSUCAL_STATUS_OPEN;
    }
  RETURN( rv );
}

/* private CSALogon() backend function */
NS_IMETHODIMP psuCalServerSunCDEImpl::CSALogon( )
{
#ifdef PSUCAL_DEBUG
  printf( "psuCalServerSunCDEImpl::CSALogon()\n" );
#endif

  nsresult rv;

  /**
   * This function is modified from sample code in:
   * "Solaris Common Desktop Envinronment: Programmers Guide"
   */

  CSA_return_code ret;
  CSA_calendar_user caddr;
  CSA_extension logon_exts[2];
  CSA_X_COM_support check_support[2];

  if (csaSessionState != CSA_SESSION_NONE)
    {
      RETURN( NS_ERROR_CAL_SERVER_ALREADY_OPEN );
    }

  /* Specify the calendar to logon to */
  caddr.user_name = NULL;
  caddr.user_type = CSA_USER_TYPE_INDIVIDUAL;
  caddr.calendar_address = nsnull;  // do not remove! buildaddress uses realloc!
  caddr.calendar_user_extensions = NULL;

  rv = BuildCSACalendarAddress( &caddr.calendar_address );
  if (NS_FAILED(rv))
    {
      RETURN( rv );
    }
  
  /* Specify the get user access extension (CSA_X_DT_GET_USER_ACCESS_EXT)
   * to retrieve the user's access right with respect to the calendar.
   */

  logon_exts[0].item_code = CSA_X_DT_GET_USER_ACCESS_EXT;
  logon_exts[0].item_data = 0;
  logon_exts[0].item_reference = NULL;
  logon_exts[0].extension_flags = 0;
  
  /* Specify the CSA_X_COM_SUPPORT_EXT extension to check
   * whether the CSA_X_XT_APP_CONTEXT_EXT extension and the
   * CSA_X_UI_ID_EXT extension are supported.
   */
  check_support[0].item_code = CSA_X_XT_APP_CONTEXT_EXT;
  check_support[0].flags = 0;
  check_support[1].item_code = CSA_X_UI_ID_EXT;
  check_support[1].flags = 0;

  logon_exts[1].item_code = CSA_X_COM_SUPPORT_EXT;
  logon_exts[1].item_data = 2;
  logon_exts[1].item_reference = NS_STATIC_CAST(CSA_buffer, check_support);
  logon_exts[0].extension_flags = CSA_EXT_LAST_ELEMENT;

  ret = csa_logon(NULL, &caddr, NULL, NULL, NULL, &csaHandle, logon_exts);

  DestroyCSACalendarAddress( caddr.calendar_address );

  if (ret == CSA_SUCCESS)
    {
      csaSessionState = CSA_SESSION_ESTABLISHED;
    }

  RETURN( CSAReturnToNSResult( ret ) );
}

/* private BuildCSACalendarAddress
 *
 * format: 'calendar@host'
 */
NS_IMETHODIMP psuCalServerSunCDEImpl::BuildCSACalendarAddress( char ** addr )
{
#ifdef PSUCAL_DEBUG
  printf( "psuCalServerSunCDEImpl::BuildCSACalendarAddress()\n" );
#endif

  if (addr == nsnull)
    {
      RETURN( NS_ERROR_NULL_POINTER );
    }
  if ((hostName == nsnull) || (calendarName == nsnull))
    {
      RETURN( NS_ERROR_CAL_SERVER_INVALID_PARAMETER );
    }

  /* size math: calendar + '@' + host + '\0' */
  *addr = NS_STATIC_CAST(char *, nsMemory::Realloc( NS_STATIC_CAST(void *, *addr), 
						    strlen( calendarName ) + 1 + 
						    strlen(hostName) + 1 ));

  if (*addr == nsnull)
    {
      RETURN( NS_ERROR_OUT_OF_MEMORY );
    }

  sprintf( *addr, "%s@%s", calendarName, hostName );

  RETURN( NS_OK );
}

/* private DestoryCSACalendarAddress */
NS_IMETHODIMP psuCalServerSunCDEImpl::DestroyCSACalendarAddress( char * addr )
{
#ifdef PSUCAL_DEBUG
  printf( "psuCalServerSunCDEImpl::DestoryCSACalendarAddress()\n" );
#endif

  if (addr == nsnull)
    {
      RETURN( NS_ERROR_NULL_POINTER );
    }
  nsMemory::Free( addr );
  RETURN( NS_OK );
}

/* void Close (); */
NS_IMETHODIMP psuCalServerSunCDEImpl::Close()
{
#ifdef PSUCAL_DEBUG
  printf( "psuCalServerSunCDEImpl::Close()\n" );
#endif

  if (connectionState != PSUCAL_STATUS_OPEN)
    {
      RETURN( NS_ERROR_CAL_SERVER_NOT_OPEN );
    }

  connectionState = PSUCAL_STATUS_AUTHENTICATED;

  RETURN( NS_OK );
}

/* void AddCalToServer (); */
/**
 *  This function is modified from sample code in:
 *   "Solaris Common Desktop Envinronment: Programmers Guide"
 */
NS_IMETHODIMP psuCalServerSunCDEImpl::AddCalToServer()
{
#ifdef PSUCAL_DEBUG
  printf( "psuCalServerSunCDEImpl::AddCalToServer()\n" );
#endif

  switch( connectionState )
    {
    case PSUCAL_STATUS_CLOSED:
    case PSUCAL_STATUS_CONNECTED:
      RETURN( NS_ERROR_CAL_SERVER_NOT_AUTHENTICATED );
      break;
    case PSUCAL_STATUS_OPEN:
      RETURN( NS_ERROR_CAL_SERVER_ALREADY_OPEN );
      break;
    default:
      break;
    }

  nsresult rv;

  CSA_return_code csa_rv;
  CSA_calendar_user caddr;
  CSA_attribute attr;
  CSA_attribute_value attr_val;

  memset( &attr_val , 0 , sizeof(CSA_attribute_value) );

  /* Specify the calendar to add */
  caddr.user_name = NULL;
  caddr.user_type = nsnull;
  caddr.calendar_address = nsnull;
  caddr.calendar_user_extensions = NULL;


  rv = BuildCSACalendarAddress( &caddr.calendar_address );
  if (NS_FAILED(rv))
    {
      RETURN( rv );
    }
#ifdef PSUCAL_DEBUG
  printf( "psuCalServerSunCDEImpl::AddCalToServer() caddr.calendar_address =[%s]\n" , caddr.calendar_address);
#endif

  /* set up the access list */
  attr_val.type = CSA_VALUE_ACCESS_LIST;
  
  rv = SetupAccessList( &attr_val.item.access_list_value );
  if( NS_FAILED(rv) )
    {
#ifdef PSUCAL_DEBUG
      printf( "Error calling psuCalServerSunCDEImpl::SetupAccessList() Error:[%x]\n", rv );
#endif
      DestroyCSACalendarAddress( caddr.calendar_address );
      DestroyAccessList(&attr_val.item.access_list_value);
      RETURN( rv );
    }

  attr.name = CSA_CAL_ATTR_ACCESS_LIST;
  attr.value = &attr_val;
  attr.attribute_extensions = NULL;

  csa_rv = csa_add_calendar(nsnull, &caddr, 1 , &attr, nsnull);
  rv = CSAReturnToNSResult(csa_rv);
  if( NS_FAILED(rv) )
    {
      DestroyCSACalendarAddress( caddr.calendar_address );
      DestroyAccessList(&attr_val.item.access_list_value);
      RETURN( rv );
    }

  rv = DestroyCSACalendarAddress( caddr.calendar_address );
  if (NS_FAILED(rv))
    {
#ifdef PSUCAL_DEBUG
      printf( "Error calling psuCalServerSunCDEImpl::DestoryCSACalendarAddress() Error:[%x]\n", rv );
#endif
      RETURN( rv );
    }

  rv = DestroyAccessList(&attr_val.item.access_list_value);
#ifdef PSUCAL_DEBUG
  if( NS_FAILED(rv) )
    {
      printf( "Error calling suCalServerSunCDEImpl::DestroyAccessList() Error:[%x]\n", rv );
    }
#endif

  RETURN( rv );
}

/* void removeCalFromServer (); */
/* This function Removes a Calendar From the server using the csa session
 *  handle.  This means that to delete a calendar, you must first use
 *  authenticate to set up the CSA_session_handle csaHandle and then 
 *  call RemoveCalFromServer to remove that calendar.  
 */
NS_IMETHODIMP psuCalServerSunCDEImpl::RemoveCalFromServer()
{
#ifdef PSUCAL_DEBUG
  printf( "psuCalServerSunCDEImpl::RemoveCalFromServer()\n" );
#endif

  switch( connectionState )
    {
    case PSUCAL_STATUS_CLOSED:
    case PSUCAL_STATUS_CONNECTED:
      RETURN( NS_ERROR_CAL_SERVER_NOT_AUTHENTICATED );
      break;
    case PSUCAL_STATUS_OPEN:
      RETURN( NS_ERROR_CAL_SERVER_ALREADY_OPEN );
      break;
    default:
      break;
    }

  nsresult rv;
  CSA_return_code csa_rv;

  /* don't use Open() here; it will create a calendar for us to delete! */
  rv = CSALogon( );
  if (NS_FAILED(rv))
    {
      RETURN( rv );
    }
  connectionState = PSUCAL_STATUS_OPEN;

  csa_rv = csa_delete_calendar( csaHandle , nsnull);
  
  rv = Close( );
#ifdef PSUCAL_DEBUG
  if (NS_FAILED(rv))
    {
      printf( "Error closing session after csa_delete_calendar() Error:[%x]\n", rv );
    }
#endif
  
  rv = CSAReturnToNSResult(csa_rv);
#ifdef PSUCAL_DEBUG  
  if( NS_FAILED(rv) )
    {
      printf( "Error calling csa_delete_calendar() Error:[%x]", rv );
    }
#endif

  RETURN( rv );
}

/* void getCalList (out unsigned long count, [array, size_is (count), retval] out string calendars); */
NS_IMETHODIMP psuCalServerSunCDEImpl::GetCalList(PRUint32 *count, char ***calendars)
{
#ifdef PSUCAL_DEBUG
  printf( "psuCalServerSunCDEImpl::GetCalList()\n" );
#endif

  switch( connectionState )
    {
    case PSUCAL_STATUS_CLOSED:
    case PSUCAL_STATUS_CONNECTED:
      RETURN( NS_ERROR_CAL_SERVER_NOT_AUTHENTICATED );
      break;
    case PSUCAL_STATUS_OPEN:
      RETURN( NS_ERROR_CAL_SERVER_ALREADY_OPEN );
      break;
    default:
      break;
    }

  if( (count == nsnull) || (calendars == nsnull) )
    RETURN( NS_ERROR_NULL_POINTER );

  CSA_return_code csa_rv;
  CSA_calendar_user *cals;
  nsresult rv;

  csa_rv = csa_list_calendars( hostName , (CSA_uint32*) count , &cals , NULL );

  rv = CSAReturnToNSResult(csa_rv);
  if (NS_FAILED(rv))
    {
#ifdef PSUCAL_DEBUG  
      printf( "Error calling csa_list_calendar() Error:[%x]\n", rv );
#endif
      RETURN( rv );
    }

  (*calendars) = (char **)nsMemory::Realloc( (*calendars) , sizeof(char *) * (*count) );
  
  if( (*calendars) == nsnull )
    {
#ifdef PSUCAL_DEBUG
      printf("psuCalServerSunCDEImpl::GetCalList() Error: Realloc returned null\n");
#endif
      RETURN( NS_ERROR_OUT_OF_MEMORY );
    }

  memset( (*calendars) , 0 , sizeof(char *) * (*count) );
  
  for(PRUint32 i = 0 ; i < *count ; i++ )
    {
      if( strlen( cals[i].calendar_address ) < strlen( "callog." ) )
	{
#ifdef PSUCAL_DEBUG
	  printf("psuCalServerSunCDEImpl::GetCalList() Error parsing calendar names\n");
#endif
	  RETURN( NS_ERROR_FAILURE );
	} 
      
      (*calendars)[i] = PL_strdup( cals[i].calendar_address + strlen("callog.") );
      if( (*calendars)[i] == nsnull )
	{
#ifdef PSUCAL_DEBUG
	  printf("psuCalServerSunCDEImpl::GetCalList() Error, out of memory\n");
#endif
	  RETURN( NS_ERROR_OUT_OF_MEMORY );
	}
    }
  
  nsMemory::Free(cals);
  
  RETURN( NS_OK );
}

/* oeIICal inherited functions */
//NS_IMPL_ISUPPORTS1(oeICal, oeIICal)

/* void Test (); */
NS_IMETHODIMP psuCalServerSunCDEImpl::Test()
{
#ifdef PSUCAL_DEBUG
  printf( "psuCalServerSunCDEImpl::Test()\n" );
#endif

  nsresult rv;
  
  rv = Connect();
  if( NS_FAILED( rv ) )
    {
#ifdef PSUCAL_DEBUG
      printf( "Error calling psuCalServerSunCDEImpl::Connect() Error:[%x]\n", rv );
#endif
      RETURN( NS_ERROR_FAILURE );
    }

  rv = Authenticate( user , "" );
  if( NS_FAILED( rv ) )
    {
#ifdef PSUCAL_DEBUG
      printf( "Error calling psuCalServerSunCDEImpl::Authenticate() Error:[%x]\n", rv );
#endif
      RETURN( NS_ERROR_FAILURE );
    }

  rv = TestGetCalList();
  if( NS_FAILED( rv ) )
    {
#ifdef PSUCAL_DEBUG
      printf( "Error calling psuCalServerSunCDEImpl::TestGetCalList() Error:[%x]\n", rv );
#endif
      RETURN( NS_ERROR_FAILURE );
    }
  
  rv = Open();
  if(NS_FAILED(rv))
    {
#ifdef PSUCAL_DEBUG
      printf( "Error calling psuCalServerSunCDEImpl::Open() Error:[%x]\n", rv );
#endif
      RETURN( NS_ERROR_FAILURE );
    }
  
  rv = TestAddEvent();  
  if( NS_FAILED( rv ) )
    {
#ifdef PSUCAL_DEBUG
      printf( "Error calling psuCalServerSunCDEImpl::TestAddEvent() Error:[%x]\n", rv );
#endif
      RETURN( NS_ERROR_FAILURE );
    }

  rv = TestSunCalEventToOeICalEvent();  
  if( NS_FAILED( rv ) )
    {
#ifdef PSUCAL_DEBUG
      printf( "Error calling psuCalServerSunCDEImpl::TestSunCalEventToOeICalEvent() Error:[%x]\n", rv );
#endif
      RETURN( NS_ERROR_FAILURE );
    }

  rv = TestModifyEvent();
  if( NS_FAILED( rv ) )
    {
#ifdef PSUCAL_DEBUG
      printf( "Error calling psuCalServerSunCDEImpl::TestModifyEvent() Error:[%x]\n", rv );
#endif
      RETURN( NS_ERROR_FAILURE );
    }

  rv = Close();
  if( NS_FAILED( rv ) )
    {
#ifdef PSUCAL_DEBUG
      printf( "Error calling psuCalServerSunCDEImpl::Close() Error:[%x]\n", rv );
#endif
      RETURN( NS_ERROR_FAILURE );
    }

  rv = Disconnect();
  if( NS_FAILED( rv ) )
    {
#ifdef PSUCAL_DEBUG
      printf( "Error calling psuCalServerSunCDEImpl::Disconnect() Error:[%x]\n", rv );
#endif
      RETURN( NS_ERROR_FAILURE );
    }

  rv = SetServer( "suncal://stewie.cede.psu.edu/testCal2" );
  if ( NS_FAILED(rv) )
    {
#ifdef PSUCAL_DEBUG
      printf("Error calling psuCalServerSunCDEImpl::SetServer() [%x]\n", rv);
#endif   
      RETURN( NS_ERROR_FAILURE );
    }
  
  rv = Connect();
  if( NS_FAILED( rv ) )
    {
#ifdef PSUCAL_DEBUG
      printf( "Error calling psuCalServerSunCDEImpl::Connect() Error:[%x]\n", rv );
#endif
      RETURN( NS_ERROR_FAILURE );
    }

  rv = Authenticate( user , "" );
  if( NS_FAILED( rv ) )
    {
#ifdef PSUCAL_DEBUG
      printf( "Error calling psuCalServerSunCDEImpl::Authenticate() Error:[%x]\n", rv );
#endif
      RETURN( NS_ERROR_FAILURE );
    }  

  rv = TestAddCal();  
  if( NS_FAILED( rv ) )
    {
#ifdef PSUCAL_DEBUG
      printf( "Error calling psuCalServerSunCDEImpl::TestAddCal() Error:[%x]\n", rv );
#endif
      RETURN( NS_ERROR_FAILURE );
    }

  rv = TestRemoveCal();  
  if( NS_FAILED( rv ) )
     {
#ifdef PSUCAL_DEBUG
       printf( "Error calling psuCalServerSunCDEImpl::TestRemoveCal() Error:[%x]\n", rv );
#endif
       RETURN( NS_ERROR_FAILURE );
     }
  
  RETURN( NS_OK );
}

/* attribute boolean batchMode; */
NS_IMETHODIMP psuCalServerSunCDEImpl::GetBatchMode(PRBool *aBatchMode)
{
#ifdef PSUCAL_DEBUG
  printf( "psuCalServerSunCDEImpl::GetBatchMode()\n" );
#endif
  if (aBatchMode == nsnull)
    {
      RETURN( NS_ERROR_NULL_POINTER );
    }

  *aBatchMode = m_batchMode;

  RETURN( NS_OK );
}

NS_IMETHODIMP psuCalServerSunCDEImpl::SetBatchMode(PRBool aBatchMode)
{
#ifdef PSUCAL_DEBUG
  printf( "psuCalServerSunCDEImpl::SetBatchMode()\n" );
#endif

  bool newBatchMode = aBatchMode;

  if( m_batchMode != newBatchMode )
    {
      m_batchMode = aBatchMode;
    }

  RETURN( NS_OK );
}

/* readonly attribute oeIICalTodo filter; */
NS_IMETHODIMP psuCalServerSunCDEImpl::GetFilter(oeIICalTodo * *aFilter)
{
#ifdef PSUCAL_DEBUG
  printf( "psuCalServerSunCDEImpl::GetFilter()\n" );
#endif

    RETURN( NS_ERROR_NOT_IMPLEMENTED );
}

/* attribute string server; */
NS_IMETHODIMP psuCalServerSunCDEImpl::GetServer(char **server) 
{               
#ifdef PSUCAL_DEBUG           
  printf( "psuCalServerSunCDEImpl::GetServer()\n" );
#endif
  if (server == nsnull)
    {
      RETURN( NS_ERROR_NULL_POINTER );
    }

  *server= (char*) nsMemory::Clone( serveraddr, strlen(serveraddr)+1 );
  if( *server == nsnull )
    {                       
      RETURN( NS_ERROR_OUT_OF_MEMORY );
    }         
  RETURN( NS_OK );
}

NS_IMETHODIMP psuCalServerSunCDEImpl::SetServer(const char *aServer)
{
#ifdef PSUCAL_DEBUG
  printf( "psuCalServerSunCDEImpl::SetServer(%s)\n", aServer != nsnull ? aServer : "(null)" );
#endif

  if (aServer == nsnull)
    {
      RETURN( NS_ERROR_NULL_POINTER );
    }

  if( strncmp( aServer, "suncal://", strlen( "suncal://" ) ) != 0 )
    {
      RETURN( NS_ERROR_INVALID_ARG );
    }

  serveraddr = (char *)nsMemory::Clone( aServer, strlen( aServer ) + 1 );
  if( serveraddr == nsnull )
    {
      RETURN( NS_ERROR_OUT_OF_MEMORY );
    }

  if (hostName != nsnull)
    {
      nsMemory::Free(hostName);
      hostName = nsnull;
    }
  if (calendarName != nsnull)
    {
      nsMemory::Free(calendarName);
      calendarName = nsnull;
    }

  RETURN( ParseSunCalURI() );
}

/* Extract hostName And calendarName */
NS_IMETHODIMP psuCalServerSunCDEImpl::ParseSunCalURI( )
{
#ifdef PSUCAL_DEBUG
  printf( "psuCalServerSunCDEImpl::ParseSunCalURI()\n" );
#endif

  char *start, *slash;
  start = serveraddr + strlen( "suncal://" );
  slash = strchr( start, '/' );

  /* string of form suncal://host */
  /* string of form suncal://host/ */
  if( slash == NULL || slash[1] == '\0' )
    {        
      RETURN( NS_ERROR_INVALID_ARG );
    } 
  /* string of form suncal://host/calendar */
  else 
    {
      hostName = (char *)nsMemory::Clone( start, slash - start + 1 );
      if( hostName == nsnull )
        {
          RETURN( NS_ERROR_OUT_OF_MEMORY );
        }
      hostName[slash - start] = '\0';

      start = slash + 1;
      calendarName = (char *)nsMemory::Clone( start, strlen( start ) + 1 );
      if( calendarName == nsnull )
        {
          RETURN( NS_ERROR_OUT_OF_MEMORY );
        }
    } 

#ifdef PSUCAL_DEBUG
  printf( "serveraddr = [%s] hostName = [%s] calendarName = [%s]\n",
          serveraddr != nsnull ? serveraddr : "(null)",
          hostName != nsnull ? hostName : "(null)",
          calendarName != nsnull ? calendarName : "(null)" );
#endif

  RETURN( NS_OK );
}

/* void addObserver (in oeIICalObserver observer); */
NS_IMETHODIMP psuCalServerSunCDEImpl::AddObserver(oeIICalObserver *observer)
{
#ifdef PSUCAL_DEBUG
  printf( "psuCalServerSunCDEImpl::AddObserver()\n" );
#endif

    RETURN( NS_ERROR_NOT_IMPLEMENTED );
}

/* void removeObserver (in oeIICalObserver observer); */
NS_IMETHODIMP psuCalServerSunCDEImpl::RemoveObserver(oeIICalObserver *observer)
{
#ifdef PSUCAL_DEBUG
  printf( "psuCalServerSunCDEImpl::RemoveObserver()\n" );
#endif

    RETURN( NS_ERROR_NOT_IMPLEMENTED );
}

/* void addTodoObserver (in oeIICalTodoObserver observer); */
NS_IMETHODIMP psuCalServerSunCDEImpl::AddTodoObserver(oeIICalTodoObserver *observer)
{
#ifdef PSUCAL_DEBUG
  printf( "psuCalServerSunCDEImpl::AddTodoObserver()\n" );
#endif

    RETURN( NS_ERROR_NOT_IMPLEMENTED );
}

/* void removeTodoObserver (in oeIICalTodoObserver observer); */
NS_IMETHODIMP psuCalServerSunCDEImpl::RemoveTodoObserver(oeIICalTodoObserver *observer)
{
#ifdef PSUCAL_DEBUG
  printf( "psuCalServerSunCDEImpl::RemoveTodoObserver()\n" );
#endif

    RETURN( NS_ERROR_NOT_IMPLEMENTED );
}

/* string addEvent (in oeIICalEvent icalevent); */
NS_IMETHODIMP psuCalServerSunCDEImpl::AddEvent(oeIICalEvent *icalevent, char **retid)
{
#ifdef PSUCAL_DEBUG
  printf( "psuCalServerSunCDEImpl::AddEvent()\n" );
#endif
  icalset *stream;
  nsresult rv;
  CSA_return_code retCode;

  if(icalevent == nsnull || retid == nsnull)
    {
      RETURN( NS_ERROR_NULL_POINTER );
    }

  icalevent->GetId(retid);
  if(*retid == nsnull)
    {
      char *id;
      rv = GenerateID(&id);
      if( NS_FAILED( rv ) )
         {
           RETURN( rv );
         }

      icalevent->SetId(id);
      icalevent->GetId(retid);
    }

  CSA_uint32 numOfAttb;
  CSA_attribute *attrs;
  CSA_entry_handle new_entry;
  /* converting entry to sun Calendar format */
  psuSunCalEvent psuEvent(icalevent, PR_TRUE);
  /* Making sure OPEN */
  if (connectionState != PSUCAL_STATUS_OPEN)
    {
#ifdef PSUCAL_DEBUG
      printf("psuCalServerSunCDEImpl::AddEvent(): ");
      printf("Error ** Connection unaviable, not in Open State\n");
#endif
      RETURN( NS_ERROR_CAL_SERVER_NOT_OPEN );
    }
  psuEvent.GetSunCalEvent(&attrs, &numOfAttb);
  /* adding sun event to calendar */
  retCode = csa_add_entry(csaHandle, 
			  numOfAttb, 
			  attrs, 
			  &new_entry, 
			  nsnull);

  
  if((retCode == CSA_SUCCESS) && (m_eventList != nsnull))
    {
      /* Adding to event list */
      m_eventList->Add(icalevent,new_entry);
    }

  RETURN( CSAReturnToNSResult(retCode) );
}

/* string modifyEvent (in oeIICalEvent icalevent); */
NS_IMETHODIMP psuCalServerSunCDEImpl::ModifyEvent(oeIICalEvent *icalevent, char **retid)
{
#ifdef PSUCAL_DEBUG
  printf( "psuCalServerSunCDEImpl::ModifyEvent()\n" );
#endif
  
  nsresult rv;
  CSA_return_code csa_rv;
  
  if(icalevent == nsnull || retid == nsnull)
    {
      RETURN( NS_ERROR_NULL_POINTER );
    }
  
  icalevent->GetId(retid);
  if(*retid == nsnull)
    {
#ifdef PSUCAL_DEBUG
      printf("psuCalServerSunCDEImpl::ModifyEvent() Error: Invalid ID\n");
#endif
      RETURN( NS_ERROR_FAILURE );
    }
  
  CSA_uint32 numOfAttb;
  CSA_attribute *attrs;
  CSA_entry_handle old_entry;
  CSA_entry_handle new_entry;
  /* converting entry to sun Calendar format */
  psuSunCalEvent psuEvent(icalevent, PR_FALSE);
  psuEvent.GetSunCalEvent( &attrs , &numOfAttb );
  
  old_entry = (*m_eventList).GetHandleById( *retid );

  /* make sure that we're in the open state */
  if (connectionState != PSUCAL_STATUS_OPEN)
    {
#ifdef PSUCAL_DEBUG
      printf("psuCalServerSunCDEImpl::ModifyEvent() Error Calendar Not Open\n");
#endif
      RETURN( NS_ERROR_CAL_SERVER_NOT_OPEN );
    }

  csa_rv = csa_update_entry_attributes( csaHandle , old_entry , CSA_SCOPE_ALL , 
  CSA_FALSE , (numOfAttb-1) , attrs , &new_entry ,
  NULL ); 
  
  rv = CSAReturnToNSResult( csa_rv );
  if( NS_FAILED( rv ) )
  {
  #ifdef PSUCAL_DEBUG
  printf( "psuCalServerSunCDEImpl::ModifyEvent() Error calling csa_update_entry_attributes()\n");
  #endif
  RETURN( rv );
  } 
  
  RETURN( NS_OK );
}

/* void deleteEvent (in string id); */
NS_IMETHODIMP psuCalServerSunCDEImpl::DeleteEvent(const char *id)
{
#ifdef PSUCAL_DEBUG
  printf( "psuCalServerSunCDEImpl::DeleteEvent()\n" );
#endif

  nsresult rv;
  CSA_return_code csa_rv;
  CSA_entry_handle entry;

  entry = (*m_eventList).GetHandleById( id );

  /* make sure that we're in the open state */
  if (connectionState != PSUCAL_STATUS_OPEN)
    {
#ifdef PSUCAL_DEBUG
      printf("psuCalServerSunCDEImpl::DeleteEvent() Error Calendar Not Open\n");
#endif
      RETURN( NS_ERROR_CAL_SERVER_NOT_OPEN );
    }

  csa_rv = csa_delete_entry( csaHandle , entry , CSA_SCOPE_ALL , NULL );

  rv = CSAReturnToNSResult( csa_rv );
  if( NS_FAILED( rv ) )
    {
#ifdef PSUCAL_DEBUG
      printf( "psuCalServerSunCDEImpl::DeleteEvent() Error calling csa_delete_entry()\n");
#endif
      RETURN( rv );
    }

  (*m_eventList).Remove( (char*) id );

  RETURN( NS_OK );
}

/* oeIICalEvent fetchEvent (in string id); */
NS_IMETHODIMP psuCalServerSunCDEImpl::FetchEvent(const char *id, 
						 oeIICalEvent **ev)
{
#ifdef PSUCAL_DEBUG
  printf( "psuCalServerSunCDEImpl::FetchEvent()\n" );
#endif
  if(id == nsnull)
    {
#ifdef ICAL_DEBUG
      printf( "psuCalServerSunCDEImpl::FetchEvent() - Invalid Id.\n" );
#endif
      *ev = nsnull;
      RETURN( NS_OK );
    }

  oeIICalEvent *event = m_eventList->GetEventById(id);
  if(event != nsnull)
    {
      event->AddRef();
    }
  *ev = event;
  RETURN( NS_OK );
}

/* nsISimpleEnumerator getAllEvents (); */
NS_IMETHODIMP psuCalServerSunCDEImpl::GetAllEvents(nsISimpleEnumerator **eventlist)
{
#ifdef PSUCAL_DEBUG
  printf( "psuCalServerSunCDEImpl::GetAllEvents()\n" );
#endif

  /** 
   *  this event turns out to be very dangerous with CSA b/c of how repeating events are handled;
   *  current idea is simply to query until the next epoch time ~2038
   */

  /* calculate the previous and next epochs (in microseconds) */
  PRTime begindate = 0;
  PRTime enddate = NS_STATIC_CAST( PRTime, 0xFFFFFFFF ) * 1000000;

  RETURN( GetNEventsForRange( begindate, enddate, PR_FALSE, 0, 0, eventlist ) );
}

/* nsISimpleEnumerator getEventsForMonth (in PRTime date); */
NS_IMETHODIMP psuCalServerSunCDEImpl::GetEventsForMonth(PRTime date, nsISimpleEnumerator **eventlist)
{
#ifdef PSUCAL_DEBUG
  printf( "psuCalServerSunCDEImpl::GetEventsForMonth()\n" );
#endif

  struct icaltimetype checkdate = ConvertFromPrtime ( date );
  checkdate.day = 1;
  checkdate.hour = 0;
  checkdate.minute = 0;
  checkdate.second = 0;
  icaltime_adjust( &checkdate, 0, 0, 0, -1 );
#ifdef PSUCAL_DEBUG
  printf( "CHECKDATE: %s\n" , icaltime_as_ctime( checkdate ) );
#endif
  PRTime checkdateinms = ConvertToPrtime( checkdate );

  struct icaltimetype checkenddate = ConvertFromPrtime ( date );
  checkenddate.month++;
  checkenddate.day = 1;
  checkenddate.hour = 0;
  checkenddate.minute = 0;
  checkenddate.second = 0;
  checkenddate = icaltime_normalize( checkenddate );
#ifdef PSUCAL_DEBUG
  printf( "CHECKENDDATE: %s\n" , icaltime_as_ctime( checkenddate ) );
#endif
  PRTime checkenddateinms = ConvertToPrtime( checkenddate );

  RETURN( GetNEventsForRange( checkdateinms, checkenddateinms,
			     PR_FALSE, 0, 0, eventlist ) );
}

/* nsISimpleEnumerator getEventsForWeek (in PRTime date); */
NS_IMETHODIMP psuCalServerSunCDEImpl::GetEventsForWeek(PRTime date, nsISimpleEnumerator **eventlist)
{
#ifdef PSUCAL_DEBUG
  printf( "psuCalServerSunCDEImpl::GetEventsForWeek()\n" );
#endif

  struct icaltimetype checkdate = ConvertFromPrtime ( date );
  checkdate.hour = 0;
  checkdate.minute = 0;
  checkdate.second = 0;
  icaltime_adjust( &checkdate, 0, 0, 0, -1 );
#ifdef PSUCAL_DEBUG
  printf( "CHECKDATE: %s\n" , icaltime_as_ctime( checkdate ) );
#endif
  PRTime checkdateinms = ConvertToPrtime( checkdate );

  struct icaltimetype checkenddate = ConvertFromPrtime ( date );
  checkenddate.hour = 0;
  checkenddate.minute = 0;
  checkenddate.second = 0;
  icaltime_adjust( &checkenddate, 7, 0, 0, 0 ); 
#ifdef PSUCAL_DEBUG
  printf( "CHECKENDDATE: %s\n" , icaltime_as_ctime( checkenddate ) );
#endif
  PRTime checkenddateinms = ConvertToPrtime( checkenddate );

  RETURN( GetNEventsForRange( checkdateinms, checkenddateinms,
			     PR_FALSE, 0, 0, eventlist ) );
}

/* nsISimpleEnumerator getEventsForDay (in PRTime date); */
NS_IMETHODIMP psuCalServerSunCDEImpl::GetEventsForDay(PRTime date, nsISimpleEnumerator **eventlist)
{
#ifdef PSUCAL_DEBUG
  printf( "psuCalServerSunCDEImpl::GetEventsForDay()\n" );
#endif

  struct icaltimetype checkdate = ConvertFromPrtime ( date );
  checkdate.hour = 0;
  checkdate.minute = 0;
  checkdate.second = 0;
  icaltime_adjust( &checkdate, 0, 0, 0, -1 );
#ifdef PSUCAL_DEBUG
  printf( "CHECKDATE: %s\n" , icaltime_as_ctime( checkdate ) );
#endif
  PRTime checkdateinms = ConvertToPrtime( checkdate );

  struct icaltimetype checkenddate = ConvertFromPrtime ( date );
  checkenddate.hour = 0;
  checkenddate.minute = 0;
  checkenddate.second = 0;
  icaltime_adjust( &checkenddate, 1, 0, 0, 0 );
#ifdef PSUCAL_DEBUG
  printf( "CHECKENDDATE: %s\n" , icaltime_as_ctime( checkenddate ) );
#endif
  PRTime checkenddateinms = ConvertToPrtime( checkenddate );

  RETURN( GetNEventsForRange( checkdateinms, checkenddateinms, 
			     PR_FALSE, 0, 0, eventlist ) );
}

/* nsISimpleEnumerator getEventsForRange (in PRTime begindate, in PRTime enddate); */
NS_IMETHODIMP psuCalServerSunCDEImpl::GetEventsForRange(PRTime begindate, PRTime enddate, nsISimpleEnumerator **eventlist)
{
#ifdef PSUCAL_DEBUG
  printf( "psuCalServerSunCDEImpl::GetEventsForRange()\n" );
#endif

  RETURN( GetNEventsForRange( begindate, enddate, PR_FALSE, 0, 0, eventlist ) );
}

/**
 *  get n events for a time range
 *
 * @param begindate (in) the beginning date of range
 * @param enddate (in) the ending date of range
 * @param limit_size (in) whether or not to limit number of returned events
 * @param count (in) ceiling on number of events returned
 * @param offset (in) first element in result set to return
 * @param events (out) the enumerator on which we return the events
 */
NS_IMETHODIMP psuCalServerSunCDEImpl::GetNEventsForRange( PRTime begindate,
							  PRTime enddate,
							  PRBool limit_size,
							  PRUint32 count,
							  PRUint32 offset,
							  nsISimpleEnumerator ** events )
{
#ifdef PSUCAL_DEBUG
  printf( "psuCalServerSunCDEImpl::GetNEventsForRange()\n" );
#endif

  CSA_uint32 num_entries;
  CSA_entry_handle * entries;

  nsresult rv;
  PRUint32 i;
  oeIICalEvent * event;
  PRUint32 limit = count;
  oeEventEnumerator * eventEnum;

  if (events == nsnull)
    {
      RETURN( NS_ERROR_NULL_POINTER );
    }
  if (*events == nsnull)
    {
      eventEnum = new oeEventEnumerator();

      if (eventEnum == nsnull)
	{
	  RETURN( NS_ERROR_OUT_OF_MEMORY );
	}

      eventEnum->QueryInterface(NS_GET_IID(nsISimpleEnumerator), (void **)events);
    }
  else
    {
      /**
       *  yuck. oeEventEnumerator really needs its own interface that 
       *  I can query to make this a safer (runtime) cast 
       */
      eventEnum = NS_STATIC_CAST( oeEventEnumerator *, *events );
    }

  rv = GetEventHandlesForRange( begindate, enddate, &num_entries, &entries );

  if (NS_FAILED( rv ))
    {
#ifdef PSUCAL_DEBUG
      printf( "psuCalServerSunCDEImpl::GetEventHandlesForRange failed; error [0x%x]\n", rv );
#endif
      RETURN( rv );
    }

  if (limit_size == PR_TRUE)
    {
      if (NS_STATIC_CAST( PRUint32, num_entries ) < count)
	{
	  limit = NS_STATIC_CAST( PRUint32, num_entries );
	}
    }
  else
    {
      offset = 0;
      limit = NS_STATIC_CAST( PRUint32, num_entries );
    }

  /* go through entry handles and get back attributes */
  for (i = offset; i < limit; i++)
    {
      event = nsnull;
      rv = RetrieveEventAttributes( entries + i, &event );
#ifdef PSUCAL_DEBUG
      if (NS_FAILED( rv ))
	{
	  printf( "psuCalServerSunCDEImpl::RetrieveEventAttributes failed; error [0x%x]\n", rv );
	}
#endif

      rv = eventEnum->AddEvent( event );
     
#ifdef PSUCAL_DEBUG
      if (NS_FAILED( rv ))
	{
	  printf( "oeEventEnumerator::AddEvent failed; error [0x%x]\n", rv );
	}
#endif

      if (m_eventList != nsnull)
	{
	  m_eventList->Add( event, entries[i] );
	}
#ifdef PSUCAL_DEBUG
      else
	{
	  printf( "uh oh...m_eventList is null!" );
	}
#endif
    }

  RETURN( NS_OK );
}

NS_IMETHODIMP psuCalServerSunCDEImpl::PRTimeToISO8601( PRTime * moztime, char ** iso8601time )
{
#ifdef PSUCAL_DEBUG
  printf( "psuCalServerSunCDEImpl::PRTimeToISO8601()\n" );
#endif

  struct tm breakdown;
  time_t epoch_time;
  nsresult rv;

  if ((moztime == nsnull) || (iso8601time == nsnull))
    {
      RETURN( NS_ERROR_NULL_POINTER );
    }

  rv = PRTimeToTM( moztime, &breakdown );
  if (NS_FAILED( rv ))
    {
      RETURN( rv );
    }

  *iso8601time = NS_STATIC_CAST( char *, nsMemory::Alloc( strlen("YYYYMMDDTHHMMSSZ") + 1 ) );
  if (*iso8601time == nsnull)
    {
      RETURN( NS_ERROR_OUT_OF_MEMORY );
    }

  sprintf( *iso8601time, "%04d%02d%02dT%02d%02d%02dZ",
	   breakdown.tm_year + 1900,
	   breakdown.tm_mon + 1,
	   breakdown.tm_mday,
	   breakdown.tm_hour,
	   breakdown.tm_min,
	   breakdown.tm_sec );

  RETURN( NS_OK );
}

/**
 *  convert PRTime to struct tm
 */
NS_IMETHODIMP psuCalServerSunCDEImpl::PRTimeToTM( PRTime * moztime, struct tm * breakdown )
{
#ifdef PSUCAL_DEBUG
  printf( "psuCalServerSunCDEImpl::PRTimeToTM()\n" );
#endif

  time_t epoch_time;

  if ((moztime == nsnull) || (breakdown == nsnull))
    {
      RETURN( NS_ERROR_NULL_POINTER );
    }

  /* convert time from microseconds since epoch to seconds since epoch */
  epoch_time = NS_STATIC_CAST( time_t, (*moztime)/1000000 );

  if (gmtime_r( &epoch_time, breakdown ) == nsnull)
    {
      RETURN( NS_ERROR_FAILURE );
    }
  RETURN( NS_OK );
}

/**
 *  convert struct tm to PRTime
 */
NS_IMETHODIMP psuCalServerSunCDEImpl::TMToPRTime( struct tm * breakdown, PRTime * moztime )
{
#ifdef PSUCAL_DEBUG
  printf( "psuCalServerSunCDEImpl::TMToPRTime()\n" );
#endif

  if ((breakdown == nsnull) || (moztime == nsnull))
    {
      RETURN( NS_ERROR_NULL_POINTER );
    }

  time_t epoch_time = mktime( breakdown );

  *moztime = NS_STATIC_CAST( PRTime, epoch_time ) * 1000000;

  RETURN( NS_OK );
}


/**
 *  build csa date filter
 */
NS_IMETHODIMP psuCalServerSunCDEImpl::BuildCSAEventFilter( CSA_attribute ** attrs, 
							   CSA_attribute_value ** attr_val, 
							   CSA_enum ** ops,
							   PRTime begindate,
							   PRTime enddate )
{
#ifdef PSUCAL_DEBUG
  printf( "psuCalServerSunCDEImpl::BuildCSADateFilter()\n" );
#endif

  if ((attrs == nsnull) || (attr_val == nsnull) || (ops == nsnull))
    {
      RETURN( NS_ERROR_NULL_POINTER );
    }

  *attrs = NS_STATIC_CAST( CSA_attribute *, nsMemory::Alloc( CSA_EVENT_FILTER_LENGTH * sizeof(CSA_attribute) ) );
  *attr_val = NS_STATIC_CAST( CSA_attribute_value *, nsMemory::Alloc( CSA_EVENT_FILTER_LENGTH * sizeof(CSA_attribute_value) ) );
  *ops = NS_STATIC_CAST( CSA_enum *, nsMemory::Alloc( CSA_EVENT_FILTER_LENGTH * sizeof(CSA_enum) ) );

  if ((*attrs == nsnull) || (*attr_val == nsnull) || (*ops == nsnull))
    {
      if (*attrs != nsnull)
	{
	  nsMemory::Free( *attrs );
	  *attrs = nsnull;
	}
      if (*attr_val != nsnull)
	{
	  nsMemory::Free( *attr_val );
	  *attr_val = nsnull;
	}
      if (*ops != nsnull)
	{
	  nsMemory::Free( *ops );
	  *ops = nsnull;
	}
      RETURN( NS_ERROR_OUT_OF_MEMORY );
    }

  int i = 0;

  /* start_date >= begindate */
  (*attrs)[i].name = CSA_ENTRY_ATTR_START_DATE;
  (*attrs)[i].value = (*attr_val) + i;
  (*attrs)[i].attribute_extensions = NULL;
  (*attr_val)[i].type = CSA_VALUE_DATE_TIME;
  PRTimeToISO8601( &begindate, &(*attr_val)[i].item.date_time_value );
  (*ops)[i] = CSA_MATCH_GREATER_THAN_OR_EQUAL_TO;
  i++;

  /* start_date < enddate */
  (*attrs)[i].name = CSA_ENTRY_ATTR_START_DATE;
  (*attrs)[i].value = (*attr_val) + i;
  (*attrs)[i].attribute_extensions = NULL;
  (*attr_val)[i].type = CSA_VALUE_DATE_TIME;
  PRTimeToISO8601( &enddate, &(*attr_val)[i].item.date_time_value );
  (*ops)[i] = CSA_MATCH_LESS_THAN;
  i++;

  /* type == event */
  (*attrs)[i].name = CSA_ENTRY_ATTR_TYPE;
  (*attrs)[i].value = (*attr_val) + i;
  (*attrs)[i].attribute_extensions = NULL;
  (*attr_val)[i].type = CSA_VALUE_UINT32;
  (*attr_val)[i].item.sint32_value = CSA_TYPE_EVENT;
  (*ops)[i] = CSA_MATCH_EQUAL_TO;
  i++;

  /* subtype == appointment */
  (*attrs)[i].name = CSA_ENTRY_ATTR_SUBTYPE;
  (*attrs)[i].value = (*attr_val) + i;
  (*attrs)[i].attribute_extensions = NULL;
  (*attr_val)[i].type = CSA_VALUE_STRING;
  (*attr_val)[i].item.string_value = CSA_SUBTYPE_APPOINTMENT;
  (*ops)[i] = CSA_MATCH_EQUAL_TO;
  i++;

  RETURN( NS_OK );
}

/**
 *  destroy csa event filter
 */
NS_IMETHODIMP psuCalServerSunCDEImpl::DestroyCSAEventFilter( CSA_attribute * attrs,
							     CSA_attribute_value * attr_val,
							     CSA_enum * ops )
{
#ifdef PSUCAL_DEBUG
  printf( "psuCalServerSunCDEImpl::DestroyCSAEventFilter()\n" );
#endif

  if ((attrs == nsnull) || (attr_val == nsnull) || (ops == nsnull))
    {
      RETURN( NS_ERROR_NULL_POINTER );
    }

  /* destroy ISO8601 strings in attr_val[0] and attr_val[1] */
  nsMemory::Free( attr_val[0].item.date_time_value );
  nsMemory::Free( attr_val[1].item.date_time_value );

  /* deallocate main filter objects */
  nsMemory::Free( attrs );
  nsMemory::Free( attr_val );
  nsMemory::Free( ops );

  RETURN( NS_OK );
}

/* NS_IMETHODIMP GetEventHandlesForRange */

NS_IMETHODIMP psuCalServerSunCDEImpl::GetEventHandlesForRange( PRTime begindate,
							       PRTime enddate,
							       CSA_uint32 * num_entries,
							       CSA_entry_handle ** entries )
{
#ifdef PSUCAL_DEBUG
  printf( "psuCalServerSunCDEImpl::GetEventHandlesForRange()\n" );
#endif

  /**
   * This function is modified from sample code in:
   * "Solaris Common Desktop Envinronment: Programmers Guide"
   */

  CSA_return_code stat;
  CSA_attribute * attrs;
  CSA_attribute_value * attr_val;
  CSA_enum * ops;

  nsresult rv;

  if ((num_entries == nsnull) || (entries == nsnull))
    {
      RETURN( NS_ERROR_NULL_POINTER );
    }

  rv = BuildCSAEventFilter( &attrs, &attr_val, &ops, begindate, enddate );

  if (NS_FAILED( rv ))
    {
#ifdef PSUCAL_DEBUG
      printf( "psuCalServerSunCDEImpl::BuldCSAEventFilter failed; error [0x%x]\n", rv );
#endif
      RETURN( NS_ERROR_FAILURE );
    }

  stat = csa_list_entries( csaHandle, CSA_EVENT_FILTER_LENGTH, attrs,
			   ops, num_entries, entries, NULL );

  rv = DestroyCSAEventFilter( attrs, attr_val, ops );

  /* DestroyCSAEventFilter failure is not fatal */
#ifdef PSUCAL_DEBUG
  if (NS_FAILED( rv ))
    {
      printf( "psuCalServerSunCDEImpl::DestroyCSAEventFilter failed; error[0x%x]\n", rv );
    }
#endif

  rv = CSAReturnToNSResult( stat );

#ifdef PSUCAL_DEBUG
  if (stat != CSA_SUCCESS)
    {
      printf( "csa_list_entries failed; ns error [0x%x]\n", rv );
    }
#endif

  RETURN( rv );
}

/**
 *  destroy event handles
 */
NS_IMETHODIMP psuCalServerSunCDEImpl::DestroyEventHandles( CSA_entry_handle * entries )
{
#ifdef PSUCAL_DEBUG
  printf( "psuCalServerSunCDEImpl::DestroyEventHandles()\n" );
#endif

  if (entries == nsnull)
    {
      RETURN( NS_ERROR_NULL_POINTER );
    }

  csa_free(entries);

  RETURN( NS_OK );
}

/**
 *  get the attributes associated with a csa event handle
 *
 * @param handle (in) the csa event handle
 * @param event (out) the oeIICalEvent object that is returned
 */
NS_IMETHODIMP psuCalServerSunCDEImpl::RetrieveEventAttributes( CSA_entry_handle * handle, oeIICalEvent ** event )
{
#ifdef PSUCAL_DEBUG
  printf( "psuCalServerSunCDEImpl::RetrieveEventAttributes()\n" );
#endif

  if ((handle == nsnull) || (event == nsnull))
    {
      RETURN( NS_ERROR_NULL_POINTER );
    }

  CSA_uint32 num_attributes;
  CSA_attribute * entry_attrs;
  CSA_return_code stat;

  nsresult rv;

  stat = csa_read_entry_attributes( csaHandle, *handle, 0, NULL, &num_attributes, &entry_attrs, NULL );

  if (stat = CSA_SUCCESS)
    {
      rv = CSAReturnToNSResult( stat );
#ifdef PSUCAL_DEBUG
      printf( "csa_read_entry_attributes failed; ns error [0x%x]\n", rv );
#endif
      RETURN( rv );
    }

  psuSunCalEvent sunEvent;
  rv = sunEvent.SetSunCalEvent( entry_attrs, num_attributes, PR_FALSE );
  if (NS_FAILED( rv ))
    {
#ifdef PSUCAL_DEBUG
      printf( "psuSunCalEvent::SetSunCalEvent failed; error [0x%x]\n", rv );
#endif
      RETURN( rv );
    }

  rv = sunEvent.SunCalEventToOeIICalEvent(event);
#ifdef PSUCAL_DEBUG
  if (NS_FAILED( rv ))
    {
      printf( "psuSunCalEvent::SunCalEventToOeIICalEvent failed; error [0x%x]\n", rv );
    }
#endif

  RETURN( rv );
}


/* nsISimpleEnumerator getNextNEvents (in PRTime begindate, in long count); */
NS_IMETHODIMP psuCalServerSunCDEImpl::GetNextNEvents(PRTime begindate, PRInt32 count, nsISimpleEnumerator **_retval)
{
#ifdef PSUCAL_DEBUG
  printf( "psuCalServerSunCDEImpl::GetNextNEvents()\n" );
#endif

    RETURN( NS_ERROR_NOT_IMPLEMENTED );
}

/* nsISimpleEnumerator getFirstEventsForRange (in PRTime begindate, in PRTime enddate); */
NS_IMETHODIMP psuCalServerSunCDEImpl::GetFirstEventsForRange(PRTime begindate, PRTime enddate, nsISimpleEnumerator **_retval)
{
#ifdef PSUCAL_DEBUG
  printf( "psuCalServerSunCDEImpl::GetFirstEventsForRange()\n" );
#endif

    RETURN( NS_ERROR_NOT_IMPLEMENTED );
}

/* string addTodo (in oeIICalTodo icaltodo); */
NS_IMETHODIMP psuCalServerSunCDEImpl::AddTodo(oeIICalTodo *icaltodo, char **_retval)
{
#ifdef PSUCAL_DEBUG
  printf( "psuCalServerSunCDEImpl::AddTodo()\n" );
#endif

    RETURN( NS_ERROR_NOT_IMPLEMENTED );
}

/* string modifyTodo (in oeIICalTodo icaltodo); */
NS_IMETHODIMP psuCalServerSunCDEImpl::ModifyTodo(oeIICalTodo *icaltodo, char **_retval)
{
#ifdef PSUCAL_DEBUG
  printf( "psuCalServerSunCDEImpl::ModifyTodo()\n" );
#endif

    RETURN( NS_ERROR_NOT_IMPLEMENTED );
}

/* void deleteTodo (in string id); */
NS_IMETHODIMP psuCalServerSunCDEImpl::DeleteTodo(const char *id)
{
#ifdef PSUCAL_DEBUG
  printf( "psuCalServerSunCDEImpl::DeleteTodo()\n" );
#endif

    RETURN( NS_ERROR_NOT_IMPLEMENTED );
}

/* oeIICalTodo fetchTodo (in string id); */
NS_IMETHODIMP psuCalServerSunCDEImpl::FetchTodo(const char *id, oeIICalTodo **_retval)
{
#ifdef PSUCAL_DEBUG
  printf( "psuCalServerSunCDEImpl::FetchTodo()\n" );
#endif

    RETURN( NS_ERROR_NOT_IMPLEMENTED );
}

/* nsISimpleEnumerator getAllTodos (); */
NS_IMETHODIMP psuCalServerSunCDEImpl::GetAllTodos(nsISimpleEnumerator **_retval)
{
#ifdef PSUCAL_DEBUG
  printf( "psuCalServerSunCDEImpl::GetAllTodos()\n" );
#endif

    RETURN( NS_ERROR_NOT_IMPLEMENTED );
}

/* void resetFilter (); */
NS_IMETHODIMP psuCalServerSunCDEImpl::ResetFilter()
{
#ifdef PSUCAL_DEBUG
  printf( "psuCalServerSunCDEImpl::ResetFilter()\n" );
#endif

    RETURN( NS_ERROR_NOT_IMPLEMENTED );
}

/**
 * Takes a SUN CSA error return code, 
 * and gives back an NS result code
 *
 * @param CSA_return_code the result code of a CSA operation
 */
nsresult psuCalServerSunCDEImpl::CSAReturnToNSResult( CSA_return_code code )
{
#ifdef PSUCAL_DEBUG
  printf( "psuCalServerSunCDEImpl::CSAReturnToNSResult(");
  if(code != CSA_SUCCESS)
    {
      printf("csa error is [%d] in decimal)\n", code );
    }
  else
    {
      printf(")\n");
    }
#endif

  nsresult ret;

  switch( code )
    {
    case CSA_SUCCESS:
      ret = NS_OK;
      break;
    case CSA_E_AMBIGUOUS_USER:
      ret = NS_ERROR_CAL_SERVER_AMBIGUOUS_USER;
      break;
    case CSA_E_CALENDAR_EXISTS:
      ret = NS_ERROR_CAL_SERVER_CALENDAR_EXISTS;
      break;
    case CSA_E_CALENDAR_NOT_EXIST:
      ret = NS_ERROR_CAL_SERVER_CALENDAR_NOT_EXIST;
      break;
    case CSA_E_CALLBACK_NOT_REGISTERED:
      ret = NS_ERROR_CAL_SERVER_CALLBACK_NOT_REGISTERED;
      break;
    case CSA_E_DISK_FULL:
      ret = NS_ERROR_CAL_SERVER_DISK_FULL;
      break;
    case CSA_E_FAILURE:
      ret = NS_ERROR_CAL_SERVER_FAILURE;
      break;
    case CSA_E_FILE_EXIST:
      ret = NS_ERROR_CAL_SERVER_FILE_EXIST;
      break;
    case CSA_E_FILE_NOT_EXIST:
      ret = NS_ERROR_CAL_SERVER_FILE_NOT_EXIST;
      break;
    case CSA_E_INSUFFICIENT_MEMORY:
      ret = NS_ERROR_OUT_OF_MEMORY;
      break;
    case CSA_E_INVALID_ATTRIBUTE:
      ret = NS_ERROR_CAL_SERVER_INVALID_ATTRIBUTE;
      break;
    case CSA_E_INVALID_ATTRIBUTE_VALUE:
      ret = NS_ERROR_CAL_SERVER_INVALID_ATTRIBUTE_VALUE;
      break;
    case CSA_E_INVALID_CALENDAR_SERVICE:
      ret = NS_ERROR_CAL_SERVER_INVALID_CALENDAR_SERVICE;
      break;
    case CSA_E_INVALID_CONFIGURATION:
      ret = NS_ERROR_CAL_SERVER_AMBIGUOUS_USER;
      break;
    case CSA_E_INVALID_DATA_EXT:
      ret = NS_ERROR_CAL_SERVER_INVALID_DATA_EXT;
      break;
    case CSA_E_INVALID_DATE_TIME:
      ret = NS_ERROR_CAL_SERVER_INVALID_DATE_TIME;
      break;
    case CSA_E_INVALID_ENTRY_HANDLE:
      ret = NS_ERROR_CAL_SERVER_INVALID_ENTRY_HANDLE;
      break;
    case CSA_E_INVALID_ENUM:
      ret = NS_ERROR_CAL_SERVER_INVALID_ENUM;
      break;
    case CSA_E_INVALID_FILE_NAME:
      ret = NS_ERROR_CAL_SERVER_INVALID_FILE_NAME;
      break;
    case CSA_E_INVALID_FLAG:
      ret = NS_ERROR_CAL_SERVER_INVALID_FLAG;
      break;
    case CSA_E_INVALID_FUNCTION_EXT:
      ret = NS_ERROR_CAL_SERVER_INVALID_FUNCTION_EXT;
      break;
    case CSA_E_INVALID_MEMORY:
      ret = NS_ERROR_CAL_SERVER_INVALID_MEMORY;
      break;
    case CSA_E_INVALID_PARAMETER:
      ret = NS_ERROR_CAL_SERVER_INVALID_PARAMETER;
      break;
    case CSA_E_INVALID_PASSWORD:
      ret = NS_ERROR_CAL_SERVER_INVALID_PASSWORD;
      break;
    case CSA_E_INVALID_RULE:
      ret = NS_ERROR_CAL_SERVER_INVALID_RULE;
      break;
    case CSA_E_INVALID_SESSION_HANDLE:
      ret = NS_ERROR_CAL_SERVER_INVALID_SESSION_HANDLE;
      break;
    case CSA_E_INVALID_USER:
      ret = NS_ERROR_CAL_SERVER_INVALID_USER;
      break;
    case CSA_E_NO_AUTHORITY:
      ret = NS_ERROR_CAL_SERVER_NO_AUTHORITY;
      break;
    case CSA_E_NOT_SUPPORTED:
      ret = NS_ERROR_CAL_SERVER_NOT_SUPPORTED;
      break;
    case CSA_E_PASSWORD_REQUIRED:
      ret = NS_ERROR_CAL_SERVER_PASSWORD_REQUIRED;
      break;
    case CSA_E_READONLY:
      ret = NS_ERROR_CAL_SERVER_READONLY;
      break;
    case CSA_E_SERVICE_UNAVAILABLE:
      ret = NS_ERROR_CAL_SERVER_SERVICE_UNAVAILABLE;
      break;
    case CSA_E_TEXT_TOO_LARGE:
      ret = NS_ERROR_CAL_SERVER_TEXT_TOO_LARGE;
      break;
    case CSA_E_TOO_MANY_USERS:
      ret = NS_ERROR_CAL_SERVER_TOO_MANY_USERS;
      break;
    case CSA_E_UNABLE_TO_OPEN_FILE:
      ret = NS_ERROR_CAL_SERVER_UNABLE_TO_OPEN_FILE;
      break;
    case CSA_E_UNSUPPORTED_ATTRIBUTE:
      ret = NS_ERROR_CAL_SERVER_UNSUPPORTED_ATTRIBUTE;
      break;
    case CSA_E_UNSUPPORTED_CHARACTER_SET:
      ret = NS_ERROR_CAL_SERVER_UNSUPPORTED_CHARACTER_SET;
      break;
    case CSA_E_UNSUPPORTED_DATA_EXT:
      ret = NS_ERROR_CAL_SERVER_UNSUPPORTED_DATA_EXT;
      break;
    case CSA_E_UNSUPPORTED_ENUM:
      ret = NS_ERROR_CAL_SERVER_UNSUPPORTED_ENUM;
      break;
    case CSA_E_UNSUPPORTED_FLAG:
      ret = NS_ERROR_CAL_SERVER_UNSUPPORTED_FLAG;
      break;
    case CSA_E_UNSUPPORTED_FUNCTION_EXT:
      ret = NS_ERROR_CAL_SERVER_UNSUPPORTED_FUNCTION_EXT;
      break;
    case CSA_E_UNSUPPORTED_PARAMETER:
      ret = NS_ERROR_CAL_SERVER_UNSUPPORTED_PARAMETER;
      break;
    case CSA_E_UNSUPPORTED_VERSION:
      ret = NS_ERROR_CAL_SERVER_UNSUPPORTED_VERSION;
      break;
    case CSA_E_USER_NOT_FOUND:
      ret = NS_ERROR_CAL_SERVER_USER_NOT_FOUND;
      break;
    case CSA_X_DT_E_BACKING_STORE_PROBLEM:
      ret = NS_ERROR_CAL_SERVER_BACKING_STORE_PROBLEM;
      break;
    case CSA_X_DT_E_ENTRY_NOT_FOUND:
      ret = NS_ERROR_CAL_SERVER_ENTRY_NOT_FOUND;
      break;
    case CSA_X_DT_E_INVALID_SERVER_LOCATION:
      ret = NS_ERROR_CAL_SERVER_INVALID_SERVER_LOCATION;
      break;
    case CSA_X_DT_E_SERVER_TIMEOUT:
      ret = NS_ERROR_CAL_SERVER_SERVER_TIMEOUT;
      break;
    case CSA_X_DT_E_SERVICE_NOT_REGISTERED:
      ret = NS_ERROR_CAL_SERVER_SERVICE_NOT_REGISTERED;
      break;
    case CSA_X_DT_E_MT_UNSAFE:
      ret = NS_ERROR_CAL_SERVER_MT_UNSAFE;
      break;
    case CSA_X_DT_E_NFS_MOUNTED_CALLOG:
      ret = NS_ERROR_CAL_SERVER_NFS_MOUNTED_CALLOG;
      break;
    default:
      ret = NS_ERROR_CAL_SERVER_FAILURE;
      break;
    }
  RETURN( ret );
}


/* End of implementation class template. */


/**********************************************************************************
*                          Other methods not in the class template                *
**********************************************************************************/
/**
 * Tests addEvent() method
 */
NS_IMETHODIMP psuCalServerSunCDEImpl::TestAddEvent()
{
#ifdef PSUCAL_DEBUG
  printf("psuCalServerSunCDEImpl::TestAddEvent()\n");
#endif
  nsresult rv;
  nsCOMPtr<oeIICalEvent> event;
  char *tmp;
  rv = CreateTestEvent(getter_AddRefs(event));
  if(rv == NS_OK)
    {
      oeIICalEvent *tmpE = event;
      rv = AddEvent(tmpE, &tmp);
    }
  RETURN( rv );
}

/**
 * Creates a Test Event to use for Testing event handling
 * @param oeIICalEvent the event to return
 */
NS_IMETHODIMP psuCalServerSunCDEImpl::CreateTestEvent(oeIICalEvent **retEvent)
{
#ifdef PSUCAL_DEBUG
  printf("psuCalServerSunCDEImpl::CreateTestEvent()\n");
#endif
  nsresult rv;
  nsCOMPtr<oeIICalEvent> testEvent = do_CreateInstance(OE_ICALEVENT_CONTRACTID, &rv);
  if(NS_FAILED(rv))
    {
      RETURN( rv );
    }
  nsCOMPtr<oeIDateTime> startTime;
  testEvent->SetTitle(NS_LITERAL_CSTRING("TEST TITLE"));
  testEvent->SetDescription(NS_LITERAL_CSTRING("TEST DESCRIPTION"));
  testEvent->SetLocation(NS_LITERAL_CSTRING("TEST LOCATION"));
  testEvent->GetStart(getter_AddRefs(startTime));
  if (NS_FAILED(startTime->SetYear(2003)) ||
      NS_FAILED(startTime->SetMonth(4)) ||
      NS_FAILED(startTime->SetDay(13)) ||
      NS_FAILED(startTime->SetHour(12)) ||
      NS_FAILED(startTime->SetMinute(12)) ||
      NS_FAILED(startTime->SetUtc(PR_FALSE)))
    {
      RETURN( NS_ERROR_FAILURE );
    }
  char *tmp;
  startTime->ToString(&tmp);
  printf("psuCalServerSunCDEImpl::TestAddEvent(): Start time is %s\n", 
	 tmp);
  tmp = nsnull;
  nsCOMPtr<oeIDateTime> endTime;
  testEvent->GetEnd(getter_AddRefs(endTime));
  if (NS_FAILED(endTime->SetYear(2003)) ||
      NS_FAILED(endTime->SetMonth(4)) ||
      NS_FAILED(endTime->SetDay(13)) ||
      NS_FAILED(endTime->SetHour(13)) ||
      NS_FAILED(endTime->SetMinute(14)) ||
      NS_FAILED(endTime->SetUtc(PR_FALSE)))
    {
      RETURN( NS_ERROR_FAILURE );
    }
  endTime->ToString(&tmp);
  printf("psuCalServerSunCDEImpl::TestAddEvent() : End time is %s\n", 
	 tmp);
  /* Setting up Email Alarm Default time unit is minute and using that */
  testEvent->SetAlarm(PR_TRUE);
  testEvent->SetAlarmEmailAddress("source@cede.psu.edu");
  testEvent->SetAlarmLength(30);
  
  testEvent->SetRecur(PR_TRUE);
  testEvent->SetRecurForever(PR_FALSE);
  testEvent->SetRecurCount(2);
  testEvent->SetRecurUnits("days");
  testEvent->SetRecurInterval(3);
  
  /* setup stuff to return the nsCOMPtr */ 
  (*retEvent) = testEvent;
  NS_ADDREF(*retEvent);
  RETURN( NS_OK );
}


/**
 * Creates a CSA_access_rights list to use when creating a calendar
 *  This function is modified from sample code in:
 *   "Solaris Common Desktop Envinronment: Programmers Guide"
 */
NS_IMETHODIMP psuCalServerSunCDEImpl::SetupAccessList(CSA_access_rights **accessList)
{
  (*accessList) = (CSA_access_rights *)nsMemory::Realloc( (*accessList) , sizeof(CSA_access_rights));
  
  if( (*accessList) == nsnull )
    {
#ifdef PSUCAL_DEBUG
      printf("psuCalServerSunCDEImpl::SetupAccessList() Error: Realloc returned null");
#endif
      RETURN( NS_ERROR_OUT_OF_MEMORY );
    }
  
  memset( *accessList , 0 , sizeof(CSA_access_rights) );
  
  (*accessList)->user = (CSA_calendar_user *)nsMemory::Realloc( (*accessList)->user , sizeof(CSA_calendar_user) );
  
  if( (*accessList)->user == nsnull )
    {
#ifdef PSUCAL_DEBUG
      printf("psuCalServerSunCDEImpl::SetupAccessList() Error: Realloc returned null");
#endif
      RETURN( NS_ERROR_OUT_OF_MEMORY );
    }
  
  (*accessList)->user->user_name = PL_strdup("world");
  (*accessList)->user->user_type = CSA_USER_TYPE_INDIVIDUAL;
  (*accessList)->rights = CSA_VIEW_PUBLIC_ENTRIES | CSA_INSERT_PUBLIC_ENTRIES;
  
  RETURN( NS_OK );
}

/**
 * Destroys a CSA_access_rights list
 *  This function is modified from sample code in:
 *   "Solaris Common Desktop Envinronment: Programmers Guide"
 */
NS_IMETHODIMP psuCalServerSunCDEImpl::DestroyAccessList(CSA_access_rights **list)
{
  CSA_access_rights *ptr;

  while((*list) != NULL)
    {
      ptr = (*list)->next;

      if((*list)->user)
	{
	  if((*list)->user->user_name)
	    nsMemory::Free((*list)->user->user_name);

	  nsMemory::Free((*list)->user);
	}

      nsMemory::Free(*list);

      (*list) = ptr;
    }

  RETURN( NS_OK );
}

/**
 * Tests AddCalToServer() method
 */
NS_IMETHODIMP psuCalServerSunCDEImpl::TestAddCal()
{
#ifdef PSUCAL_DEBUG
  printf("psuCalServerSunCDEImpl::TestAddCal()\n");
#endif

  nsresult rv;

  rv = AddCalToServer();
  if( NS_FAILED(rv) )
    {
#ifdef PSUCAL_DEBUG
      printf( "Error calling psuCalServerSunCDEImpl::AddCalToServer() Error:[%x]\n", rv );
#endif      
      RETURN( NS_ERROR_FAILURE );
    }
  
  RETURN( NS_OK );
}

/**
 * Tests RemoveCalFromServer() method
 */
NS_IMETHODIMP psuCalServerSunCDEImpl::TestRemoveCal()
{
#ifdef PSUCAL_DEBUG
  printf("psuCalServerSunCDEImpl::TestRemoveCal()\n");
#endif

  nsresult rv;

  rv = RemoveCalFromServer();
  if( NS_FAILED( rv ) )
    {
#ifdef PSUCAL_DEBUG
      printf( "Error calling psuCalServerSunCDEImpl::RemoveCalFromServer() Error:[%x]\n", rv );
#endif
      RETURN( NS_ERROR_FAILURE );
    }

  RETURN( NS_OK );
}

/**
 * Tests GetCalList() method
 */
NS_IMETHODIMP psuCalServerSunCDEImpl::TestGetCalList()
{
#ifdef PSUCAL_DEBUG
  printf("psuCalServerSunCDEImpl::TestGetCalList()\n");
#endif

  nsresult rv;
  
  PRUint32 count;
  char **calendars;

  rv = GetCalList( &count , &calendars );
  if( NS_FAILED( rv ) )
    {
#ifdef PSUCAL_DEBUG
      printf("Error calling psuCalServerSunCDEImpl::GetCalList() Error:[%x]\n",rv);
#endif
      RETURN( rv );
    }

  for(PRUint32 i = 0 ; i < count ; i++)
    {
      printf("[%d]:%s\n",i,calendars[i]);
    }

  nsMemory::Free(calendars);

  RETURN( NS_OK );
}


NS_IMETHODIMP psuCalServerSunCDEImpl::TestSunCalEventToOeICalEvent()
{
#ifdef PSUCAL_DEBUG
  printf("psuCalServerSunCDEImpl::TestSunCalEventToOeICalEvent()\n");
#endif

  nsresult rv;
  CSA_return_code stat;
  CSA_attribute attrs[4];
  CSA_attribute_value attr_val[4];
  CSA_enum ops[4];
  CSA_uint32 i;
  CSA_uint32 num_entries;
  CSA_entry_handle *entries;
  CSA_uint32 num_attributes;
  CSA_attribute *entry_attrs;

  i = 0;

  attrs[i].name = CSA_ENTRY_ATTR_START_DATE;
  attrs[i].value = &attr_val[i];
  attrs[i].attribute_extensions = NULL;
  attr_val[i].type = CSA_VALUE_DATE_TIME;
  attr_val[i].item.date_time_value = "19960801T000000Z";
  ops[i++] = CSA_MATCH_GREATER_THAN_OR_EQUAL_TO;

  attrs[i].name = CSA_ENTRY_ATTR_START_DATE;
  attrs[i].value = &attr_val[i];
  attrs[i].attribute_extensions = NULL;
  attr_val[i].type = CSA_VALUE_DATE_TIME;
  attr_val[i].item.date_time_value = "20060801T000000Z";
  ops[i++] = CSA_MATCH_LESS_THAN_OR_EQUAL_TO;

  oeIICalEvent *event;

  stat = csa_list_entries(csaHandle, i, attrs, ops, &num_entries, &entries, NULL);
  if(stat != CSA_SUCCESS)
    {
      RETURN( CSAReturnToNSResult(stat) );
    }

  printf("TestSunCalEventToOeICalEvent(): got num_entries = [%d]\n", num_entries);
  for(i = 0; i < num_entries; i++ )
    {
      stat = csa_read_entry_attributes( csaHandle, entries[i], 0, NULL, &num_attributes, &entry_attrs, NULL );
      if(stat != CSA_SUCCESS)
        {
          csa_free( entry_attrs );
          printf( "psuCalServerSunCDEImpl::TestSunCalEventToOeICalEvent(): csa_read_entry_attrs() failed with in i = [%d]\n", i );
          RETURN( CSAReturnToNSResult( stat ) );
        }
      printf( "TestSunCalEventToOeICalEvent(): got num_attributes = [%d] in i = [%d]\n", num_entries, i );
      psuSunCalEvent se( entry_attrs, num_attributes, PR_FALSE );
      rv = se.GetOeIICalEvent( &event );
      if( NS_FAILED( rv ) )
        {
          printf( "psuCalServerSunCDEImpl::TestSunCalEventToOeICalEvent(): GetOeIICalEvent failed in i = [%d] with ns err [%X]\n", i, rv ); 
          RETURN( rv );
        }
      event->Release();
    }

  RETURN( NS_OK );
}

NS_IMETHODIMP psuCalServerSunCDEImpl::GenerateID(char **id)
{
#ifdef PSUCAL_DEBUG
  printf("psuCalServerSunCDEImpl::GenerateID()\n");
#endif
  unsigned long newid;

  if ( id == nsnull )
    {
      RETURN( NS_ERROR_NULL_POINTER );
    }

  *id = (char *)nsMemory::Alloc( 10 );
  if (*id == nsnull)
    {
      RETURN( NS_ERROR_OUT_OF_MEMORY );
    }

  newid = 900000000 + ((rand()%0x4ff) << 16) + rand()%0xFFFF;
  sprintf( *id, "%lu", newid );

  RETURN( NS_OK );
}

NS_IMETHODIMP psuCalServerSunCDEImpl::TestSetServer()
{
#ifdef PSUCAL_DEBUG
  printf("psuCalServerSunCDEImpl::TestSetSErver()\n");
#endif

  /* These first 2 should fail */
  nsresult rv = SetServer( "suncal://stewie.cede.psu.edu" );
  if ( NS_FAILED(rv) )
    {
      printf("ERROR: Calling psuCalServerSunCDEImpl::SetServer() [%X]\n", rv);
    }

  rv = SetServer( "suncal://stewie.cede.psu.edu/" );
  if ( NS_FAILED(rv) )
    {
      printf("ERROR: Calling psuCalServerSunCDEImpl::SetServer() [%X]\n", rv);
    }

  rv = SetServer( "suncal://stewie.cede.psu.edu/foobar" );
  if ( NS_FAILED(rv) )
    {
      printf("ERROR: Calling psuCalServerSunCDEImpl::SetServer() [%X]\n", rv);
    }

  RETURN( rv );
}


NS_IMETHODIMP psuCalServerSunCDEImpl::TestEventList()
{
#ifdef PSUCAL_DEBUG
  printf("psuCalServerSunCDEImpl::TestEventList()\n");
#endif

  EventList *el = new EventList();
  delete el;

  RETURN( NS_ERROR_NOT_IMPLEMENTED );
}

EventList::EventList()
{
  event = nsnull;
  sunEventHandle = nsnull;
  next = nsnull;
}

EventList::~EventList()
{
  if( event )
    event->Release();
  if( sunEventHandle )
    csa_free( (void *)sunEventHandle );
  if( next )
    delete next;
}

void EventList::Add( oeIICalEvent* e, CSA_entry_handle h )
{
  if( !event )
    {
      event = e;    
      NS_ADDREF(event);
      sunEventHandle = h;
    }
  else
    {
      if( !next )
        {
          next = new EventList();
        }
      next->Add( e, h );
    }
}

oeIICalEvent* EventList::GetEventById( const char *id )
{
  if( !event )
    return nsnull;
  if( MatchId( id ) )
    return event;
  if( next )
    return next->GetEventById( id );
  return nsnull;
}

CSA_entry_handle EventList::GetHandleById( const char *id )
{
  if( !sunEventHandle || !event )
    return nsnull;
  if( MatchId( id ) )
    return sunEventHandle;
  if( next )
    return next->GetHandleById( id );
  return nsnull;
}

void EventList::Remove( char *id )
{
  if( !event )
    return;
  if( MatchId( id ) ) 
    {
      if( event )
        event->Release();
      if( sunEventHandle )
        csa_free( (void *)sunEventHandle );
      if( next )
        {
          event = next->event;
          EventList *tmp = next;
          next = next->next;
          tmp->next = nsnull;
          tmp->event = nsnull;
          tmp->sunEventHandle = nsnull;
          delete tmp;
        }
      else
        {
          event = nsnull;
          sunEventHandle = nsnull;
          id = nsnull;
        }
    }
  else
    {
      if( next )
        next->Remove( id );
    }
}

PRBool EventList::MatchId( const char *id )
{
  char *thisId;
  PRBool result;
  if ( id == nsnull || event == nsnull )
    {
      return PR_FALSE;
    }

  (void)event->GetId( &thisId );

  result = strcmp( id, thisId ) == 0 ? PR_TRUE : PR_FALSE;
  nsMemory::Free( thisId );
  return result;
}

/**
 * Tests modifyEvent() method
 */
NS_IMETHODIMP psuCalServerSunCDEImpl::TestModifyEvent()
{
#ifdef PSUCAL_DEBUG
  printf("psuCalServerSunCDEImpl::TestModifyEvent()\n");
#endif
  nsresult rv;
  nsCOMPtr<oeIICalEvent> event;
  char *tmp;

  rv = CreateTestEvent(getter_AddRefs(event));
  if(NS_FAILED ( rv ) )
    {
#ifdef PSUCAL_DEBUG
      printf("psuCalServerSunCDEImpl::TestModifyEvent() Error Calling CreateTestEvent\n");
#endif
      RETURN( rv );
    }

  rv = AddEvent(event, &tmp);
  if( NS_FAILED( rv ) )
    {
#ifdef PSUCAL_DEBUG
      printf("psuCalServerSunCDEImpl::TestModifyEvent() Error Calling Addevent\n");
#endif
      RETURN( rv );
    }

  event->SetTitle(NS_LITERAL_CSTRING("MODIFIED TEST TITLE"));

  rv = ModifyEvent( event , &tmp );

  RETURN( rv );
}

/**********************************************************************************
 *                          Other functions from oeICal                           *
**********************************************************************************/

static icaltimetype ConvertFromPrtime( PRTime indate ) 
{
  icaltimetype outdate = icaltime_null_time();
  
  PRExplodedTime ext;
  PRInt64 indateinusec, usecpermsec;
  LL_I2L( usecpermsec, PR_USEC_PER_MSEC );
  LL_MUL( indateinusec, indate, usecpermsec );
  PR_ExplodeTime( indateinusec, PR_LocalTimeParameters, &ext);
  
  outdate.year = ext.tm_year;
  outdate.month = ext.tm_month + 1;
  outdate.day = ext.tm_mday;
  outdate.hour = ext.tm_hour;
  outdate.minute = ext.tm_min;
  outdate.second = ext.tm_sec;

  //Assume that daylight time saving was not in effect before 1970
  if( ext.tm_year <= 1969 && ext.tm_params.tp_dst_offset == 3600 )
    icaltime_adjust( &outdate, 0, -1, 0, 0 );

  return outdate;
}

static PRTime ConvertToPrtime ( icaltimetype indate ) 
{
  PRExplodedTime ext;
  ext.tm_year = indate.year;
  ext.tm_month = indate.month - 1;
  ext.tm_mday = indate.day;
  ext.tm_hour = indate.hour;
  ext.tm_min = indate.minute;
  ext.tm_sec = indate.second;
  ext.tm_usec = 0;
  ext.tm_params.tp_gmt_offset = 0;
  ext.tm_params.tp_dst_offset = 0;
  
  PRTime result = PR_ImplodeTime( &ext );
  PR_ExplodeTime( result, PR_LocalTimeParameters, &ext);
  
  //Assume that daylight time saving was not in effect before 1970
  if( ext.tm_year <= 1969 && ext.tm_params.tp_dst_offset == 3600 )
    ext.tm_params.tp_dst_offset = 0;

  ext.tm_year = indate.year;
  ext.tm_month = indate.month - 1;
  ext.tm_mday = indate.day;
  ext.tm_hour = indate.hour;
  ext.tm_min = indate.minute;
  ext.tm_sec = indate.second;
  ext.tm_usec = 0;
  result = PR_ImplodeTime( &ext );
  PRInt64 usecpermsec;
  LL_I2L( usecpermsec, PR_USEC_PER_MSEC );
  LL_DIV( result, result, usecpermsec );
  return result;
}

