/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is PSU Calendar Server Code, released March 3rd, 2003.
 *
 * The Initial Developer of the Original Code is
 * Penn State CSE 420w Mozilla Project
 * Portions created by the Initial Developer are Copyright (C) 2003
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):  Daniel Bungert <drb@debian.org>
 *                  Thomas Keiser <tkeiser@cede.psu.edu>
 *                  Benjamin Nave <bnave@psu.edu>
 *                  John Pendzick <jnp117@psu.edu>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */
#ifndef _PSU_CAL_SERVER_SUN_CDE_IMPL_H_
#define _PSU_CAL_SERVER_SUN_CDE_IMPL_H_

#include "psuICalServer.h"
#include "stdlib.h"
#include "psuSunCalEvent.h"

// uuid 08113718-e3b1-4139-9034-0d413d548b35
#define PSU_SUNCDE_CID {0x08113718, 0xe3b1, 0x4139, 0x90, 0x34, 0x0d, 0x41, 0x3d, 0x54, 0x8b, 0x35}

#define PSU_SUNCDE_CONTRACTID "@mozilla.org/ical/calserver;1"

extern "C" {
  #include "csa/csa.h"
}

/* defines current length of CSA event filter */
#define CSA_EVENT_FILTER_LENGTH 4

/***********      oeICal Functions     *****************/
static icaltimetype ConvertFromPrtime( PRTime indate );
static PRTime ConvertToPrtime ( icaltimetype indate );

/***********    end oeICal Functions   *****************/


class EventList
{
public:
    oeIICalEvent *event;
    CSA_entry_handle sunEventHandle;
    EventList *next;

    EventList();
    ~EventList();
    void Add( oeIICalEvent *e, CSA_entry_handle h );
    oeIICalEvent *GetEventById( const char *id );
    CSA_entry_handle GetHandleById( const char *id );
    void Remove( char *id );
    PRBool MatchId( const char *id );
};

class psuCalServerSunCDEImpl : public psuICalServer
{
 public:
  NS_DECL_ISUPPORTS
  NS_DECL_OEIICAL
  NS_DECL_PSUICALSERVER

  psuCalServerSunCDEImpl();
  virtual ~psuCalServerSunCDEImpl();

  /**
   * Generate the identifier used in list comparison function operator=
   *
   * @param Generated identifier
   */
  NS_IMETHODIMP static GenerateID(char **id);

 private:
  /**
   *  CSA session state
   */
  typedef enum
    {
      CSA_SESSION_NONE,
      CSA_SESSION_ESTABLISHED,
      CSA_SESSION_BROKEN
    } csa_session_state_t;
  csa_session_state_t csaSessionState;

  /**
   *  convert CSA_return_code to nsresult
   */
  static nsresult CSAReturnToNSResult( CSA_return_code );

  /**
   *  build csa server address
   */
  NS_IMETHODIMP BuildCSACalendarAddress( char ** );

  /**
   *  destory csa server address
   */
  NS_IMETHODIMP DestroyCSACalendarAddress( char * );

  /**
   *  backend function to perform csa_logon
   */
  NS_IMETHODIMP CSALogon( );

  /**
   *  convert PRTime to ISO8601 string
   */
  static NS_IMETHODIMP PRTimeToISO8601( PRTime *, char ** );

  /**
   *  convert PRTime to struct tm
   */
  static NS_IMETHODIMP PRTimeToTM( PRTime *, struct tm * );

  /**
   *  convert struct tm to PRTime
   */
  static NS_IMETHODIMP TMToPRTime( struct tm *, PRTime * );

  /**
   *  build csa event filter
   *
   * @param attrs (out) pointer to CSA_attribute *
   * @param attr_val (out) pointer to a CSA_attribute_value *
   * @param ops (out) pointer to a CSA_enum *
   * @param begindate (in) the start date for the filter
   * @param enddate (in) the end date for the filter
   */
  static NS_IMETHODIMP BuildCSAEventFilter( CSA_attribute ** attrs, 
					    CSA_attribute_value ** attr_val, 
					    CSA_enum ** ops,
					    PRTime begindate,
					    PRTime enddate );

  /**
   *  destory csa event filter
   *
   * @param attrs (in) pointer CSA_attribute vector
   * @param attr_val (in) pointer to CSA_attribute_value vector
   * @param ops (in) pointer to CSA_enum vector
   */
  static NS_IMETHODIMP DestroyCSAEventFilter( CSA_attribute * attrs,
					      CSA_attribute_value * attr_val,
					      CSA_enum * ops );

  /**
   *  get the attributes associated with a csa event handle
   *
   * @param handle (in) the csa event handle
   * @param event (out) the oeIICalEvent object that is returned
   */
  NS_IMETHODIMP RetrieveEventAttributes( CSA_entry_handle * handle, oeIICalEvent ** event );

  /**
   *  get event handles for a time range
   */
  NS_IMETHODIMP GetEventHandlesForRange( PRTime begindate,
					 PRTime enddate,
					 CSA_uint32 * num_entries,
					 CSA_entry_handle ** entries );
				    
  /**
   *  destroy event handles
   */
  NS_IMETHODIMP DestroyEventHandles( CSA_entry_handle * entries );

  /**
   *  get n events for a time range
   *
   * @param begindate the beginning of event return date range
   * @param enddate the end of event return date range
   * @param limit_size whether to limit number of returned events
   * @param count maximum number of events to return
   * @param offset first event to return
   * @param events enumeration of returned events
   */
  NS_IMETHODIMP GetNEventsForRange( PRTime begindate,
				    PRTime enddate,
				    PRBool limit_size,
				    PRUint32 count,
				    PRUint32 offset,
				    nsISimpleEnumerator ** events );

  /**
   *  libcsa handle
   */
  CSA_session_handle csaHandle;

  /**
   *  server hostname
   */
  char *hostName;

  /**
   *  server-side identifier of calendar
   */
  char *calendarName;

  /**
   *  user id on server
   */
  char *user;

  /**
   *  password on server
   */
  char *passwd;

  /**
   *  connection state enumeration
   */
  serverStatusProperty connectionState;

  /**
   *  oeIICalEvent <-> SunCalEvent proxy/storage/retrivial
   */
  EventList *m_eventList;

  /**
   *  unparsed server string, in format suncal://host/calendar
   */
  char *serveraddr;

  /**
   *  indicate if event retrival is on or off.
   */
  bool m_batchMode;

  /**
   * Test for the addEvent method
   *
   */
  NS_IMETHODIMP TestAddEvent();

  /**
   * Creates a CSA_access_rights list to use when creating a calendar
   *  This function is modified from sample code in:
   *   "Solaris Common Desktop Envinronment: Programmers Guide"
   */
  NS_IMETHODIMP SetupAccessList(CSA_access_rights **accessList);

  /**
   * Destroys a CSA_access_rights list
   *  This function is modified from sample code in:
   *   "Solaris Common Desktop Envinronment: Programmers Guide"
   */
  NS_IMETHODIMP DestroyAccessList(CSA_access_rights **list);

  /**
   * suncal uri parser - extract and assign hostName and calendarName from the uri
   */
  NS_IMETHODIMP ParseSunCalURI( );
  
  /**
   * Tests AddCalToServer() method
   */
  NS_IMETHODIMP TestAddCal();
    
  /**
   * Tests RemoveCalFromServer() method
   */
  NS_IMETHODIMP TestRemoveCal();

  /**
   * Tests GetCalList() method
   */
  NS_IMETHODIMP TestGetCalList();

  /**
   * Creates a Test Event to use for Testing event handling
   * @param oeIICalEvent the event to return
   */
  NS_IMETHODIMP CreateTestEvent(oeIICalEvent **);

  /**
   * Test all member functions of EventList 
   */
  NS_IMETHODIMP TestEventList();

  /**
   * Test SetServer() method
   */
  NS_IMETHODIMP TestSetServer();

  /**
   * Test conversion from attrList / numAttr to oeICalEvent
   */
  NS_IMETHODIMP TestSunCalEventToOeICalEvent();

  /**
   * Tests ModifyEvent()
   */
  NS_IMETHODIMP TestModifyEvent();

};


#endif
