/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is PSU Calendar Server Code, released March 3rd, 2003.
 *
 * The Initial Developer of the Original Code is
 * Penn State CSE 420w Mozilla Project
 * Portions created by the Initial Developer are Copyright (C) 2003
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):  Daniel Bungert <drb@debian.org>
 *                  Thomas Keiser <tkeiser@cede.psu.edu>
 *                  Benjamin Nave <bnave@psu.edu>
 *                  John Pendzick <jnp117@psu.edu>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */
#ifndef _PSU_SUN_CALL_EVENT_CLASS_H_
#define _PSU_SUN_CALL_EVENT_CLASS_H_

//#include "psuCalServerSunCDEImpl.h"
#include "nsMemory.h"
#include "nsCOMPtr.h"
#include "nsError.h"
#include "nsString.h"

#include "psuICalServerError.h"
#include "nsComponentManagerUtils.h"
#include <stdlib.h>

#include "oeICalEventImpl.h"

extern "C" {
#include "ical.h"
#include "icalss.h"
#include "csa/csa.h"
}

#ifdef PSUCAL_DEBUG 
#define RETURN( RV ) \
        do \
          { \
            nsresult debugrv = RV; \
            if( NS_FAILED(debugrv) ) \
              { \
                printf( __FILE__ ":%d returning [0x%x]\n", __LINE__, debugrv ); \
              } \
            return debugrv; \
          } \
        while ( 0 )
#else
#define RETURN( RV ) do { return RV; } while ( 0 )
#endif

/* The date used by suncsa server to indicate that the event repeats (or go on until) forever */
#define FOREVER_DATE "19700101T000000Z"

/* repeat information is gathered up here; after we're done processing attributes,
 * AssignRepeatInformation() actually fills oeICalEvent with this information */
class RepeatInfo
{
public:
    PRUint32 type;
    PRUint32 times;
    PRUint32 interval;
    char * isodate;
    PRInt32 occurnum;
    RepeatInfo()  { type = times = interval = occurnum = 0; isodate = nsnull; }
    ~RepeatInfo() { if ( isodate != nsnull ) nsMemory::Free( isodate ); }
};

/** Max number of Sun Calendar Attributes that a Sun Calendar Event can have */
#define SUNCAL_MAX_ATTRIBE 9

class psuSunCalEvent
{
 private:
  CSA_entry_handle m_Entry;
  CSA_attribute *m_AttrList;
  CSA_uint32 m_NumAttr;
  PRBool m_MemAttr;
  PRBool m_Addable;
  /**
   *  Deletes the Attribute that object currently also resets m_NumAttr back
   *  to zero
   */
  NS_IMETHODIMP csaEntryAttrDelete();

  /**
   * Converts a oeIICalEvent to SunCalEvent
   * @param oeIICalEvent the oeIICalEvent to covert into a Sun CDE calendar event
   * @param  PRBool is the event being added?
   */
  NS_IMETHODIMP OeIICalEventToSunCalEvent(oeIICalEvent *, PRBool);
  
  /**
   * 
   * @event the oeIICalEvent that will make the string out off
   * @param retString return the string
   * @param the event to get the Description from
   */
  NS_IMETHODIMP MakeDescription(char **, oeIICalEvent *);
  
  /**
   *  Creates the CSA_reminder attribute for mail out of the 
   *  oeIICalEvent handed to it
   *  @param CSA_reminder the return CSA_reminder that will be email reminder
   *  @param oeIICalEvent the oeIICalEvent to build the reminder with
   */
  NS_IMETHODIMP CreateEmailReminder(CSA_reminder *, oeIICalEvent*);
  
  /**
   * Creates the Recurence String in the format that CSA is wanting
   * @param char** the return string
   * @param  oeIICalEvent* the event to make into a  repeating string
   */
  NS_IMETHODIMP CreateRecurenceString(char **retString, 
				      oeIICalEvent *event);
  /**
   * Initialization function
   */
  NS_IMETHODIMP init();

  /**
   *  Handles Deleteing the attribute value
   *  @param CSA_attribute_value* the attribute to delete
   */
  NS_IMETHODIMP CSAFreeAttrValue(CSA_attribute_value *);

  /**
   * Free the CSA_extension stuff
   * @param CSA_extension* the extension to free
   */
  NS_IMETHODIMP CSAFreeExtensions(CSA_extension *);
  
  /**
   *  Free the reminder attribute given
   * @param CSA_reminder* the reminder attribute to free
   */
  NS_IMETHODIMP CSAFreeReminder(CSA_reminder *);
  
  /**
   * Free the opaque Data given
   * @param CSA_opaque_data* the opaque data to free
   */
  NS_IMETHODIMP CSAFreeOpaqueData(CSA_opaque_data *);

  /**
   * Given a RepeatInfo and an oeIICalEvent, this assigns the repeat information
   * from the RepeatInfo to the oeIICalEvent 
   * @param oeIICalEvent event to fill with repeat information
   * @param RepeatInfo source of repeat info
   */
  NS_IMETHODIMP AssignRepeatInformation( oeIICalEvent *event, RepeatInfo *repeat );
 
   /**
    * assign classification status (Public/Private) to event
    * @param CSA_attribute_value raw csa data containing classification
    * @param oeIICalEvent event to assign classification
    */
   NS_IMETHODIMP DoAttrClassification( CSA_attribute_value *value, oeIICalEvent *event );

  /**
   * assign creation date to event
   * @param CSA_attribute_value raw csa data containing date
   * @param oeIICalEvent event to assign creation date
   */
  NS_IMETHODIMP DoAttrDateCreated( CSA_attribute_value *value, oeIICalEvent *event );

  /**
   * assign end date to event
   * @param CSA_attribute_value raw csa data containing date
   * @param oeIICalEvent event to assign end date
   */
  NS_IMETHODIMP DoAttrEndDate( CSA_attribute_value *value, oeIICalEvent *event );

  /**
   * establish an email reminder
   * @param CSA_attribute_value raw csa data containg email address and when to send reminder
   * @param oeIICalEvent event to assign reminder
   */
  NS_IMETHODIMP DoAttrMailReminder( CSA_attribute_value *value, oeIICalEvent *event );

  /**
   * extract repeat interval
   * @param CSA_attribute_value raw csa data containg repeat interval
   * @param RepeatInfo storage for repeat interval
   */
  NS_IMETHODIMP DoAttrRepeatInterval( CSA_attribute_value *value, RepeatInfo *r );

  /**
   * extract repeat type
   * @param CSA_attribute_value raw csa data containg repeat type
   * @param RepeatInfo storage for repeat type
   */
  NS_IMETHODIMP DoAttrRepeatType( CSA_attribute_value *value, RepeatInfo *r );

  /**
   * extract how many times event repeats
   * @param CSA_attribute_value raw csa data containg repeat times
   * @param RepeatInfo storage for repeat times
   */
  NS_IMETHODIMP DoAttrRepeatTimes( CSA_attribute_value *value, RepeatInfo *r );

  /**
   * extract which location in recurrence this event is
   * @param CSA_attribute_value raw csa data containg occurence number
   * @param RepeatInfo storage for occurence number
   */
  NS_IMETHODIMP DoAttrRepeatOccurrenceNum( CSA_attribute_value *value, RepeatInfo *r );

  /**
   * extract when sequence ends
   * @param CSA_attribute_value raw csa data containg sequence end date
   * @param RepeatInfo storage for sequence end date
   */
  NS_IMETHODIMP DoAttrRepeatSequenceEndDate( CSA_attribute_value *value, RepeatInfo *r );

  /**
   * assign start date to event
   * @param CSA_attribute_value raw csa data containing start date
   * @param oeIICalEvent event to assign start date
   */
  NS_IMETHODIMP DoAttrStartDate( CSA_attribute_value *value, oeIICalEvent *event );

  /**
   * assign summary to event
   * @param CSA_attribute_value raw csa data containing summary
   * @param oeIICalEvent event to assign summary
   */
  NS_IMETHODIMP DoAttrSummary( CSA_attribute_value *value, oeIICalEvent *event );

  /**
   * get type of entry
   * @param CSA_attribute_value raw csa data containing type
   * @param oeIICalEvent event to assign type
   */
  NS_IMETHODIMP DoAttrType( CSA_attribute_value *value, oeIICalEvent *event );

 public:
  /** Constructors **/
  psuSunCalEvent();
  psuSunCalEvent(oeIICalEvent*, PRBool);
  psuSunCalEvent(CSA_attribute *, CSA_uint32, PRBool);
  
  /** Desturctor **/
  ~psuSunCalEvent();
  
  /**
   * Get the OeIICalEvent that is in this class
   * @param oeIICalEvent** the return oeIICalEvent
   */
  NS_IMETHODIMP GetOeIICalEvent(oeIICalEvent **);


  /**
   * Set the SunCal Event handle for this event
   * @param CSA_entry_handle the event handle that it will be set to
   */
  NS_IMETHODIMP SetSunCalHandle(CSA_entry_handle);
  
  /**
   * Returns the SunCal Event Handle
   * @param CSA_entry_handle** the returning event handle
   */
  NS_IMETHODIMP GetSunCalHandle(CSA_entry_handle*);

  /**
   *  Get the SunCalEvent for this object
   *  @param CSA_attribute **const constant returning attribute
   *  @param CSA_uint32 *const return number of attributes for this event
   */
  NS_IMETHODIMP GetSunCalEvent(CSA_attribute ** const, CSA_uint32* const);
  
  /**
   * Set the Object to hold this SunCal Event
   * @param CSA_attribute * the attribute that this event should hold
   * @param CSA_uint32 the number of attributes in this SunCal Event
   */
  NS_IMETHODIMP SetSunCalEvent(CSA_attribute *, CSA_uint32, PRBool);

  /**
   * Set the SunCal Event by an oeIICalEvent
   * @param oeIICalEvent* the oeIICalEvent to make into a SunCalEvent
   *                      and hold on to
   * @param PRBool should this SunCalevent be able to be added
   */
  NS_IMETHODIMP SetSunCalEvent(oeIICalEvent *, PRBool);

  /**
   * Converts this psuSunCalEvent to oeIICalEvent
   *
   * @param oeIICalEvent the resultant oeIICalEvent
   */
  NS_IMETHODIMP SunCalEventToOeIICalEvent(oeIICalEvent **);
  
  /**
   * Acquire lead time from string of form "%cPT%dS"
   * where %c is + for a reminder before the event, and - for after the event
   * and %d is an unsigned integer - oe only supports leading reminders, so we
   * do too
   *
   * @param PRUint32 extracted lead time, in seconds
   * @param char * string to extract lead time from
   */
  NS_IMETHODIMP ExtractLeadTime(PRUint32 *, char *);
  
  /**
   *
   * @param the oeIDateTime with time values from string filled in
   * @param Iso8601 formatted string
   */
  static NS_IMETHODIMP Iso8601ToOeDateTime(oeIDateTime*, char *);

  /**
   *
   * returning oeIDateTime as a ISO-8601 formated string
   * YYYY-MM-DDThh:mm:SSTZD the string actually is for csa "YYYYMMDDTHHMMSSZ"
   * @param char the return string in iso8601 format
   * @param oeIDateTime the oeIDateTime to convert to iso8601
   */
  static NS_IMETHODIMP OeDateTimeToIso8601(char **,  oeIDateTime *);
};

#endif


